#!/usr/bin/env python
# coding: utf-8
# +
"""
This module tests if uploading data in chunks works as
expected (we should get the correct DataFrame length back).
"""
import pandas as pd
import pytest

# local imports
from pangres import aupsert, upsert
from pangres.examples import _TestsExampleTable
from pangres.tests.conftest import adrop_table_between_tests, drop_table_between_tests, sync_or_async_test, TableNames


# -

# # Sync and async variants for tests
#
# (`run_test_foo`|`run_test_foo_async`) -> `test_foo`

# +
@drop_table_between_tests(table_name=TableNames.VARIOUS_CHUNKSIZES)
def run_test_various_chunksizes(engine, schema, chunksize, nb_rows):
    # get example df
    df = _TestsExampleTable.create_example_df(nb_rows=nb_rows)

    # MySQL does not want flexible text length in indices/PK
    upsert(con=engine, schema=schema, table_name=TableNames.VARIOUS_CHUNKSIZES,
           df=df, chunksize=chunksize, if_row_exists='update')
    df_db = _TestsExampleTable.read_from_db(engine=engine, schema=schema, table_name=TableNames.VARIOUS_CHUNKSIZES)

    # sort index (for MySQL...)
    pd.testing.assert_frame_equal(df.sort_index(), df_db.sort_index())


@adrop_table_between_tests(table_name=TableNames.VARIOUS_CHUNKSIZES)
async def run_test_various_chunksizes_async(engine, schema, chunksize, nb_rows):
    # get example df
    df = _TestsExampleTable.create_example_df(nb_rows=nb_rows)

    # MySQL does not want flexible text length in indices/PK
    await aupsert(con=engine, schema=schema, table_name=TableNames.VARIOUS_CHUNKSIZES,
                  df=df, chunksize=chunksize, if_row_exists='update')
    df_db = await _TestsExampleTable.aread_from_db(engine=engine, schema=schema,
                                                   table_name=TableNames.VARIOUS_CHUNKSIZES)

    # sort index (for MySQL...)
    pd.testing.assert_frame_equal(df.sort_index(), df_db.sort_index())


# -

# # Actual tests

@pytest.mark.parametrize('chunksize, nb_rows', [[1, 11], [3, 11]], ids=['one_by_one', 'odd_chunksize'])
def test_insert_various_chunksizes(engine, schema, chunksize, nb_rows):
    sync_or_async_test(engine=engine, schema=schema,
                       f_async=run_test_various_chunksizes_async,
                       f_sync=run_test_various_chunksizes,
                       chunksize=chunksize,
                       nb_rows=nb_rows)



import pandas as pd
import pytest
from pangres.utils import adjust_chunksize, fix_psycopg2_bad_cols
from pangres.exceptions import (DuplicateLabelsException,
                                TooManyColumnsForUpsertException,
                                UnnamedIndexLevelsException)


# # Tests

# +
def test_unnamed_ix_levels_pg_column_names_fix(_):
    # make a df that has no name for its index
    df = pd.DataFrame({'id': [0]})
    with pytest.raises(UnnamedIndexLevelsException) as exc_info:
        fix_psycopg2_bad_cols(df=df)
    assert 'index levels must be named' in str(exc_info.value)


def test_duplicates_before_pg_column_names_fix(_):
    """
    Tests the case when the user passes a df with duplicated labels
    to `fix_psycopg2_bad_cols`
    """
    # make a df where "id" is both a column name and an index label
    df = pd.DataFrame({'id': [0]}).set_index('id')
    df['id'] = 0
    with pytest.raises(DuplicateLabelsException) as exc_info:
        fix_psycopg2_bad_cols(df=df)
    assert 'There cannot be duplicated' in str(exc_info.value)


def test_duplicates_after_pg_column_names_fix(_):
    """
    Tests the case when after cleaning the labels of a df
    with `fix_psycopg2_bad_cols` we end up with duplicated labels
    """
    # "test(" and "test)" will both be cleaned to "test" resulting in duplicate labels
    df = pd.DataFrame({'id': [0], 'test(': [0], 'test)': [0]}).set_index('id')
    with pytest.raises(DuplicateLabelsException) as exc_info:
        fix_psycopg2_bad_cols(df=df)
    assert 'not unique after renaming' in str(exc_info.value)


params_bad_replacements = [{'@': ''}, False, None, {'(': 1, ')': 2, '%': 3}]


@pytest.mark.parametrize('replacements', params_bad_replacements, ids=[str(p) for p in params_bad_replacements])
def test_bad_replacements_for_pg_column_names_fix(_, replacements):
    """
    When the user gives a wrong object type or wrong replacements
    for `fix_psycopg2_bad_cols` (e.g. the user tries to replace
    the character "@" although this does not cause any problems with psycopg2)
    """
    df = pd.DataFrame({'id': [0]}).set_index('id')
    with pytest.raises(TypeError) as exc_info:
        fix_psycopg2_bad_cols(df=df, replacements=replacements)
    e_str = str(exc_info.value)
    assert 'must be a dict' in e_str or 'values of replacements must all be strings' in e_str


@pytest.mark.parametrize('nb_columns', [2, 10, 50_000])
def test_adjust_chunksize(engine, schema, nb_columns):
    chunksize = 100_000  # make this very high on purpose so that it gets reduced
    # IMPORTANT: -1 because the index level counts as a column
    df = (pd.DataFrame({i: [0] for i in range(nb_columns - 1)})
          .rename_axis(index='profileid'))
    # currently we only know of restrictions for SQlite and asyncpg
    has_limitations = any(s in engine.dialect.dialect_description for s in ('sqlite', 'asyncpg'))
    if not has_limitations:
        pytest.skip('no known limitations for the number of SQL parameters in a statement for this engine')

    # this is for not repeating ourselves too much
    test_func = lambda: adjust_chunksize(con=engine, df=df, chunksize=chunksize)

    if nb_columns == 50_000:
        with pytest.raises(TooManyColumnsForUpsertException) as exc_info:
            test_func()
        assert 'allowed parameters' in str(exc_info)
    else:
        # we may need to implement different expectations depending on dialect someday
        # but currently in the case of SQlite or asyncpg we end up with the same number
        expectations = {2: 16383, 10: 3276}  # {nb_columns:expected_chunksize}
        assert test_func() == expectations[nb_columns]



#!/usr/bin/env python
# coding: utf-8
# +
"""
This module contains the functions of pangres
that will be directly exposed to its users.
"""
import pandas as pd
from sqlalchemy.engine import Connectable
from typing import Union

# local imports
from pangres.executor import Executor
from pangres.helpers import validate_chunksize_param
from pangres.pangres_types import AsyncConnectable, AUpsertResult, UpsertResult


# -

# # upsert

def upsert(con: Connectable,
           df: pd.DataFrame,
           table_name: str,
           if_row_exists: str,
           schema: Union[str, None] = None,
           create_schema: bool = False,
           create_table: bool = True,
           add_new_columns: bool = False,
           adapt_dtype_of_empty_db_columns: bool = False,
           chunksize: Union[int, None] = None,
           dtype: Union[dict, None] = None,
           yield_chunks: bool = False) -> UpsertResult:
    """
    Insert updates/ignores a pandas DataFrame into a SQL table (or
    creates a SQL table from the DataFrame if it does not exist).

    This index of the DataFrame must be the primary key or a unique key
    (can be a unique constraint with multiple columns too) of the SQL
    table and meet the following conditions:
    1. The index must be unique.
    2. All levels of the index have to be named.
    3. The index cannot contain any null value except for MySQL where
    rows with null in indices will be skipped and warnings will be raised.

    **GOTCHAS**:

    Please head over to https://github.com/ThibTrip/pangres/#Gotchas-and-caveats
    or read pangres/README.md.

    **IMPORTANT NOTES ON CONNECTIONS AND TRANSACTIONS**:

    * If an Engine object (sqlalchemy.engine.base.Engine) is given for the parameter `con`:

    We will create a connection and a transaction and handle everything (commit|rollback
    and closing both the connection and transaction).

    * If a connection object (sqlalchemy.engine.base.Connection) is given for the parameter `con`:

    We will not close the connection (so that you may reuse it later). You will have to close
    it yourself!

    We will also not create or handle transactions. This is for allowing users to make commit-as-you-go
    workflows.

    Note that on **sqlalchemy >= 2.0** (or in 1.4 when passing `future=True` when creating an engine) sqlalchemy
    will implicitely create a transaction requiring you to **commit** pangres' operations when using
    a Connection (see Examples)!

    For examples using transactions and commit-as-you-go strategies, see this notebook:
    https://github.com/ThibTrip/pangres/blob/master/demos/transaction_control.ipynb

    Notes
    -----
    It is recommanded to use this function with big batches of data
    as there is quite the overhead. Setting the arguments create_schema,
    add_new_columns and adapt_dtype_of_empty_db_columns to False should
    drastically reduce the overhead if you do not need such features.

    Parameters
    ----------
    con : sqlalchemy.engine.base.Engine or sqlalchemy.engine.base.Connection
        See https://docs.sqlalchemy.org/en/latest/core/engines.html
    df : pd.DataFrame
        See https://pandas.pydata.org/pandas-docs/stable/user_guide/10min.html
    table_name : str
        Name of the SQL table
    if_row_exists : {'ignore', 'update'}
        Behavior if a row exists in the SQL table.
        Irrelevant if the SQL table does not exist already.
        For both 'ignore' and 'update' rows with new primary/unique
        key values are inserted.

        * 'ignore': rows for which primary/unique keys exist in the SQL
        table are skipped
        * 'update': rows for which primary/unique keys exist in the SQL
        table are updated (with the columns available in the pandas
        DataFrame)
    schema : str or None, default None
        Name of the SQL schema that contains or will contain the table
        For postgres if it is None it will default to "public".
        For MySQL and SQlite the schema should be None since
        those SQL flavors do not have this system.
    create_schema : bool, default False
        If True the schema is created if it does not exist
    create_table : bool, default True
        If True the table is created if it does not exist
    add_new_columns : bool, default False
        If True adds columns present in the DataFrame that
        are not in the SQL table.
    adapt_dtype_of_empty_db_columns : bool, default False
        If True looks for columns that have no data in the
        SQL table but have data in the DataFrame;
        if those columns have datatypes that do not match
        (e.g. "TEXT" in the SQL table and "int64" in the DataFrame)
        then they are altered in the SQL table.
        Data type conversion must be supported by the SQL flavor!
        E.g. for Postgres converting from BOOLEAN to TIMESTAMP
        will not work even if the column is empty.
    chunksize : int or None, default None
        Specify the number of rows in each batch to be written at a time.
        By default, all rows will be written at once.
    dtype : None or dict {str:SQL_TYPE}, default None
        Similar to pd.to_sql dtype argument.
        This is especially useful for MySQL where the length of
        primary keys with text has to be provided (see Examples)
    yield_chunks : bool, default False
        If True gives back an sqlalchemy object
        (sqlalchemy.engine.cursor.LegacyCursorResult)
        at each chunk with which you can for instance count rows.

    Raises
    ------
    pangres.exceptions.HasNoSchemaSystemException
        When `create_schema` is True but the SQL flavor of given
        engine has no schema system (AFAIK only PostgreSQL has a
        schema system)
    pangres.exceptions.BadColumnNamesException
        When column names are incompatible with the SQL driver
        (e.g. psycopg2 does not accept "(" in a column name)
    pangres.exceptions.UnnamedIndexLevelsException
        When some of the index levels of the df are not named
    pangres.exceptions.DuplicateValuesInIndexException
        When any entry in the df's index is duplicated
    pangres.exceptions.DuplicateLabelsException
        When any name across the df's index/columns appears more than once
    pangres.exceptions.MissingIndexLevelInSqlException
        When `add_new_columns` is True but one of the columns missing
        in the SQL table is actually part of the df's index

    Examples
    --------
    #### 1. Workflow example

    ##### 1.1. Creating a SQL table
    >>> import pandas as pd
    >>> from pangres import upsert, DocsExampleTable
    >>> from sqlalchemy import create_engine, text, VARCHAR
    >>>
    >>> # create a SQLalchemy engine
    >>> engine = create_engine("sqlite://")
    >>>
    >>> # this is necessary if you want to test with MySQL
    >>> # instead of SQlite or Postgres because MySQL needs
    >>> # to have a definite limit for text primary keys/indices
    >>> dtype = {'full_name':VARCHAR(50)}
    >>>
    >>> # get or create a pandas DataFrame
    >>> # for our example full_name is the index
    >>> # and will thus be used as primary key
    >>> df = DocsExampleTable.df
    >>> print(df.to_markdown()) # to_markdown exists since pandas v1
    | full_name     | likes_sport   | updated                   |   size_in_meters |
    |:--------------|:--------------|:--------------------------|-----------------:|
    | John Rambo    | True          | 2020-02-01 00:00:00+00:00 |             1.77 |
    | The Rock      | True          | 2020-04-01 00:00:00+00:00 |             1.96 |
    | John Travolta | False         | NaT                       |           nan    |

    >>> # create SQL table
    >>> # it does not matter if if_row_exists is set
    >>> # to "update" or "ignore" for table creation
    >>> upsert(con=engine,
    ...        df=df,
    ...        table_name='example',
    ...        if_row_exists='update',
    ...        dtype=dtype)

    >>> # alternative for the statement above using a connection
    >>> # instead of an engine (the same logic can be applied for
    >>> # all usages of the `upsert` function below)
    >>> with engine.connect() as con:
    ...     upsert(con=con, df=df, table_name='example',
    ...            if_row_exists='update', dtype=dtype)
    ...     # con.commit() # IMPORTANT! uncomment if using sqlalchemy >= 2.0

    ##### 1.2. Updating the SQL table we created with if_row_exists='update'
    >>> new_df = DocsExampleTable.new_df
    >>> print(new_df.to_markdown())
    | full_name             | likes_sport   | updated                   |   size_in_meters |
    |:----------------------|:--------------|:--------------------------|-----------------:|
    | John Travolta         | True          | 2020-04-04 00:00:00+00:00 |             1.88 |
    | Arnold Schwarzenegger | True          | NaT                       |             1.88 |

    >>> # insert update using our new data
    >>> # we know the table already exists so we can pass `create_table=False`
    >>> # to skip the table existence check and gain a little time
    >>> # note that if the table did not exist we would get an error!
    >>> upsert(con=engine,
    ...        df=new_df,
    ...        table_name='example',
    ...        if_row_exists='update',
    ...        dtype=dtype,
    ...        create_table=False)
    >>>
    >>> # Now we read from the database to check what we got and as you can see
    >>> # John Travolta was updated and Arnold Schwarzenegger was added!
    >>> with engine.connect() as connection:
    ...     query = text('SELECT * FROM example')
    ...     print(pd.read_sql(query, con=connection, index_col='full_name')
    ...           .astype({'likes_sport':bool}).to_markdown())
    | full_name             | likes_sport   | updated                    |   size_in_meters |
    |:----------------------|:--------------|:---------------------------|-----------------:|
    | John Rambo            | True          | 2020-02-01 00:00:00.000000 |             1.77 |
    | The Rock              | True          | 2020-04-01 00:00:00.000000 |             1.96 |
    | John Travolta         | True          | 2020-04-04 00:00:00.000000 |             1.88 |
    | Arnold Schwarzenegger | True          |                            |             1.88 |

    ##### 1.3. Updating the SQL table with if_row_exists='ignore'
    >>> new_df2 = DocsExampleTable.new_df2
    >>> print(new_df2.to_markdown())
    | full_name     | likes_sport   | updated   |   size_in_meters |
    |:--------------|:--------------|:----------|-----------------:|
    | John Travolta | True          | NaT       |             2.5  |
    | John Cena     | True          | NaT       |             1.84 |

    >>> upsert(con=engine,
    ...        df=new_df2,
    ...        table_name='example',
    ...        if_row_exists='ignore',
    ...        dtype=dtype,
    ...        create_table=False)
    >>>
    >>>
    >>> # Now we read from the database to check what we got and as you can see
    >>> # John Travolta was NOT updated and John Cena was added!
    >>> with engine.connect() as connection:
    ...     query = text('SELECT * FROM example')
    ...     print(pd.read_sql(query, con=connection, index_col='full_name')
    ...           .astype({'likes_sport':bool}).to_markdown())
    | full_name             | likes_sport   | updated                    |   size_in_meters |
    |:----------------------|:--------------|:---------------------------|-----------------:|
    | John Rambo            | True          | 2020-02-01 00:00:00.000000 |             1.77 |
    | The Rock              | True          | 2020-04-01 00:00:00.000000 |             1.96 |
    | John Travolta         | True          | 2020-04-04 00:00:00.000000 |             1.88 |
    | Arnold Schwarzenegger | True          |                            |             1.88 |
    | John Cena             | True          |                            |             1.84 |

    #### 2. Example for getting information on upserted chunks (parameter `yield_chunks` == True)
    >>> import pandas as pd
    >>> from pangres import upsert, DocsExampleTable
    >>> from sqlalchemy import create_engine, VARCHAR
    >>>
    >>> # config
    >>> engine = create_engine("sqlite://")
    >>> chunksize = 2
    >>>
    >>> # get a DataFrame from somwhere
    >>> df = DocsExampleTable.df
    >>> print(df.to_markdown())
    | full_name     | likes_sport   | updated                   |   size_in_meters |
    |:--------------|:--------------|:--------------------------|-----------------:|
    | John Rambo    | True          | 2020-02-01 00:00:00+00:00 |             1.77 |
    | The Rock      | True          | 2020-04-01 00:00:00+00:00 |             1.96 |
    | John Travolta | False         | NaT                       |           nan    |

    {{raw}}
    >>> # upsert in chunks of size `chunksize` and get
    >>> # back the number of rows updated for each chunk
    >>> iterator = upsert(con=engine, df=df, table_name='test_row_count',
    ...                   chunksize=chunksize, if_row_exists='update',
    ...                   yield_chunks=True)
    >>> for result in iterator:
    ...     print(f'{result.rowcount} row(s) updated')
    2 row(s) updated
    1 row(s) updated
    """
    # verify arguments
    if if_row_exists not in ('ignore', 'update'):
        raise ValueError('if_row_exists must be "ignore" or "update"')
    if chunksize is None:
        chunksize = len(df)  # we'll attempt to insert the whole df at once
    else:
        validate_chunksize_param(chunksize=chunksize)

    # create object that will execute all SQL operations
    executor = Executor(df=df, table_name=table_name, schema=schema, create_schema=create_schema,
                        create_table=create_table, dtype=dtype, add_new_columns=add_new_columns,
                        adapt_dtype_of_empty_db_columns=adapt_dtype_of_empty_db_columns)

    # execute SQL operations
    if not yield_chunks:
        executor.execute(connectable=con, if_row_exists=if_row_exists, chunksize=chunksize)
        return None
    else:
        return executor.execute_yield(connectable=con, if_row_exists=if_row_exists, chunksize=chunksize)


# # Async upsert

async def aupsert(con: AsyncConnectable,
                  df: pd.DataFrame,
                  table_name: str,
                  if_row_exists: str,
                  schema: Union[str, None] = None,
                  create_schema: bool = False,
                  create_table: bool = True,
                  add_new_columns: bool = False,
                  adapt_dtype_of_empty_db_columns: bool = False,
                  chunksize: Union[int, None] = None,
                  dtype: Union[dict, None] = None,
                  yield_chunks: bool = False) -> AUpsertResult:
    """
    Asynchronous variant of `pangres.upsert`. Make sure to read its docstring
    before using this function!

    The parameters of `pangres.aupsert` are the same but parameter `con`
    will require an asynchronous connectable (asynchronous engine or asynchronous connection).

    For example, you can use PostgreSQL asynchronously with `sqlalchemy` thanks to
    the library/driver `asyncpg`, or SQLite with `aiosqlite` or Mysql with `aiomysql`.

    **WARNING**

    Although this has been well tested in `pangres` you should read this warning from
    `sqlalchemy` (underlying library we use):

    > The asyncio extension as of SQLAlchemy 1.4.3 can now be considered to be beta level software.
    > API details are subject to change however at this point it is unlikely for
    > there to be significant backwards-incompatible changes.
    >
    > Source: https://docs.sqlalchemy.org/en/14/orm/extensions/asyncio.html

    **IMPORTANT NOTES ON CONNECTIONS AND TRANSACTIONS**:

    The notes on transactions in the docstring `pangres.upsert` apply here as well
    but async connections do **not autocommit** even prior to sqlalchemy 2.0.
    This means that if you are passing an asynchronous connection you will have to commit yourself
    (manually via the connection or a transaction or context managers..., see Examples below).

    **IMPORTANT info for IPYTHON/JUPYTER users**

    You will need to run the following code before executing asynchronous code in an IPython context
    (this includes Jupyter notebooks and also inside of Jupyter Lab). This is not specific to `pangres`.

    ```
    import nest_asyncio  # pip install nest_asyncio
    nest_asyncio.apply()
    ```

    Not doing this could result in the following exception occuring:

    ```
    RuntimeError: asyncio.run() cannot be called from a running event loop
    ```

    **IMPORTANT info about parallelism**

    `aupsert` is subject to race conditions due to its asynchronous nature. For instance,
    imagine the following scenario:

    1. 2 coroutines (parallel functions in Python) A and B both do an upsert operation
    2. A and B check if the table does not exist at roughly the same time
    3. The database tells both coroutines that the table does not exist
    4. Both coroutines try to create the table at roughly the same time:
       one coroutine should succeed but the other one will raise an Exception

    For examples of what kind of race conditions can occur, see this notebook:
    https://github.com/ThibTrip/pangres/blob/master/demos/gotchas_asynchronous_pangres.ipynb

    Examples
    --------
    >>> import asyncio
    >>> import pandas as pd
    >>> from pangres import aupsert, DocsExampleTable
    >>> from sqlalchemy.ext.asyncio import create_async_engine # doctest: +SKIP
    >>>
    >>> # config
    >>> engine = create_async_engine("postgresql+asyncpg://username:password@localhost:5432/postgres") # doctest: +SKIP
    >>>
    >>> # get some data
    >>> df = DocsExampleTable.df
    >>>
    >>> # Create table before inserting! This will avoid race conditions mentionned above
    >>> # (here we are lazy, so we'll use pangres to do that, but we could also use a sqlalchemy ORM model)
    >>> # By using `df.head(0)` we get 0 rows, but we have all the information about columns, index levels
    >>> # and data types that we need for creating the table.
    >>> # And in a second step (see coroutine `execute_upsert` that we define after)
    >>> # we will set all parameters that could cause structure changes
    >>> # to False, so we can run queries in parallel without worries!
    >>> async def setup():
    ...     await aupsert(con=engine, df=df.head(0),
    ...                   table_name='example',
    ...                   if_row_exists='update',
    ...                   create_schema=True,
    ...                   create_table=True,
    ...                   add_new_columns=True)
    >>> asyncio.run(setup()) # doctest: +SKIP
    >>>
    >>> # now that we know the table exists, let's insert data into it
    >>> # the example is a bit stupid since we only have one coroutine but
    >>> # you should get the idea.
    >>> # See variable `coroutines` below where we could add several coroutines
    >>> # in order to make queries in parallel!)
    >>> async def execute_upsert():
    ...     async with engine.connect() as connection:
    ...         await aupsert(con=connection, df=df,
    ...                       table_name='example',
    ...                       if_row_exists='update',
    ...                       # set this to False (other structure related parameters are False by default)
    ...                       create_table=False)
    ...         await connection.commit()  # !IMPORTANT
    >>> loop = asyncio.get_event_loop() # doctest: +SKIP
    >>> coroutines = [execute_upsert()] # doctest: +SKIP
    >>> tasks = asyncio.gather(*coroutines) # doctest: +SKIP
    >>> loop.run_until_complete(tasks) # doctest: +SKIP

    * alternative to the coroutine above but iterating over results of inserted chunks
      so that we can for instance get the number of updated rows
    >>> async def execute_upsert():
    ...     async with engine.connect() as connection:
    ...         # this creates an async generator
    ...         async_gen = await aupsert(con=connection, df=df,
    ...                                   table_name='example',
    ...                                   if_row_exists='update',
    ...                                   yield_chunks=True,  # <--- WE SET THIS TO TRUE for iterating
    ...                                   # set this to False (other structure related parameters are False by default)
    ...                                   create_table=False)
    ...         async for result in async_gen:
    ...             print(f'{result.rowcount} updated rows')  # print the number of updated rows
    ...
    ...         # !IMPORTANT (note that you could also commit between chunks if that makes any sense in your case)
    ...         await connection.commit()
    """
    # verify arguments
    if if_row_exists not in ('ignore', 'update'):
        raise ValueError('if_row_exists must be "ignore" or "update"')
    if chunksize is None:
        chunksize = len(df)  # we'll attempt to insert the whole df at once
    else:
        validate_chunksize_param(chunksize=chunksize)

    # create object that will execute all SQL operations
    executor = Executor(df=df, table_name=table_name, schema=schema, create_schema=create_schema,
                        create_table=create_table, dtype=dtype, add_new_columns=add_new_columns,
                        adapt_dtype_of_empty_db_columns=adapt_dtype_of_empty_db_columns)

    # execute SQL operations
    if not yield_chunks:
        await executor.aexecute(async_connectable=con, if_row_exists=if_row_exists, chunksize=chunksize)
        return None
    else:
        # IMPORTANT! NO `await` because this returns an asynchronous generator
        return executor.aexecute_yield(async_connectable=con, if_row_exists=if_row_exists, chunksize=chunksize)



# +
class HasNoSchemaSystemException(Exception):
    """
    Raised when trying or preparing to execute operations
    related to SQL schemas on a non-compatible database
    (AFAIK schemas are a postgreSQL only feature)

    E.g. in `pangres.upsert` when `create_schema=True`
    and the database is a SQLite database.
    """
    pass


class BadColumnNamesException(Exception):
    """
    Raised when a user passes a DataFrame with column
    names incompatible with the SQL driver.

    E.g. psycopg2 (driver for PostgreSQL) cannot insert
    columns with parentheses in the name
    See https://github.com/psycopg/psycopg2/issues/167
    """
    pass


class UnnamedIndexLevelsException(Exception):
    """
    Raised in `pangres.upsert` when the user
    passes a DataFrame where some index levels
    have no names
    """
    pass


class DuplicateValuesInIndexException(Exception):
    """
    Raised in `pangres.upsert` when the user
    passes a DataFrame where some index entries
    are duplicated
    """
    pass


class DuplicateLabelsException(Exception):
    """
    Raised in `pangres.upsert` when the user
    passes a DataFrame where a name appears more than once
    across columns and index levels
    """
    pass


class MissingIndexLevelInSqlException(Exception):
    """
    Raised when the user asks to add a column into a SQL table
    but the column is part of the index we are supposed to use
    to upsert data
    """
    pass


class TooManyColumnsForUpsertException(Exception):
    """
    Raised when a DataFrame has so many columns, that an upsert
    operation of even one row cannot succeed due to limitations
    in the maximum number of SQL parameters of given database
    """
    pass



# +
import logging
import pandas as pd
from math import floor
from sqlalchemy.engine import Connectable

# local imports
from pangres.helpers import _sqlite_gt3_32_0, validate_chunksize_param
from pangres.logger import log
from pangres.exceptions import (DuplicateLabelsException,
                                TooManyColumnsForUpsertException,
                                UnnamedIndexLevelsException)


# -

# # Function to fix bad column names for psycopg2

def fix_psycopg2_bad_cols(df: pd.DataFrame, replacements: dict = {'%': '', '(': '', ')': ''}) -> pd.DataFrame:
    """
    Replaces '%', '(' and ')' (characters that won't play nicely or even
    at all with psycopg2) in column and index names in a deep copy of df.
    This is a workaround for the unresolved issue
    described here: https://github.com/psycopg/psycopg2/issues/167

    **IMPORTANT**:
    You will need to apply the same changes in the database as
    well if the SQL table already exists for a given DataFrame.
    Otherwise you will for instance end up with a column
    "total_%" and "total_" in your SQL table.

    Parameters
    ----------
    df : pd.DataFrame
    replacements : dict {'%':str, '(':str, ')':str}, default {'%':'', '(':'', ')':''}
        The keys '%', '(' and ')' are mandatory.
        There cannot be any extra keys.

    Raises
    ------
    pangres.exceptions.UnnamedIndexLevelsException
        When you pass a df where not all index levels are named
    pangres.exceptions.DuplicateLabelsException
        When you pass a df with duplicated labels accross index/columns
        or when after cleaning we end up with duplicated labels
        e.g. "test(" and "test)" would by default both be renamed to "test"
    TypeError
        When `replacements` is not of the expected type or has wrong keys
        or has non string values

    Returns
    -------
    new_df : pd.DataFrame

    Examples
    --------
    * fix bad col/index names with default replacements (empty string for '(', ')' and '%')
    >>> from pangres import fix_psycopg2_bad_cols
    >>> import pandas as pd
    >>> df = pd.DataFrame({'test()':[0],
    ...                    'foo()%':[0]}).set_index('test()')
    >>> print(df.to_markdown())
    |   test() |   foo()% |
    |---------:|---------:|
    |        0 |        0 |

    >>> df_fixed = fix_psycopg2_bad_cols(df)
    >>> print(df_fixed.to_markdown())
    |   test |   foo |
    |-------:|------:|
    |      0 |     0 |

    * fix bad col/index names with custom replacements - you MUST provide replacements for '(', ')' and '%'!
    >>> import pandas as pd
    >>> df = pd.DataFrame({'test()':[0],
    ...                    'foo()%':[0]}).set_index('test()')
    >>> print(df.to_markdown())
    |   test() |   foo()% |
    |---------:|---------:|
    |        0 |        0 |

    >>> df_fixed = fix_psycopg2_bad_cols(df, replacements={'%':'percent', '(':'', ')':''})
    >>> print(df_fixed.to_markdown())
    |   test |   foopercent |
    |-------:|-------------:|
    |      0 |            0 |
    """
    # verify all index levels are named
    index_names = list(df.index.names)
    if any([ix_name is None for ix_name in index_names]):
        raise UnnamedIndexLevelsException("All index levels must be named!")

    # verify duplicated columns
    fields = list(df.index.names) + df.columns.tolist()
    if len(set(fields)) != len(fields):
        duplicates = [c for c in fields if fields.count(c) > 1]
        raise DuplicateLabelsException("There cannot be duplicated names amongst "
                                       f"index levels and/or columns! Duplicates found: {duplicates}")
    # verify replacements arg
    expected_keys = ('%', '(', ')')
    not_a_dict = not isinstance(replacements, dict)
    not_all_keys_present = not_a_dict or set(replacements.keys()) - set(expected_keys) != set()
    bad_nb_keys = not_a_dict or len(replacements) != len(expected_keys)
    if not_all_keys_present or bad_nb_keys:
        raise TypeError(f'replacements must be a dict containing the following keys (and none other): {expected_keys}')
    if not all((isinstance(v, str) for v in replacements.values())):
        raise TypeError('The values of replacements must all be strings')

    # replace bad col names
    translator = {ord(k): v for k, v in replacements.items()}
    new_df = df.copy(deep=True)
    renamer = lambda col: col.translate(translator) if isinstance(col, str) else col
    new_df = new_df.rename(columns=renamer).rename_axis(index=renamer)

    # check columns are unique after renaming
    fields = list(new_df.index.names) + new_df.columns.tolist()
    if len(set(fields)) != len(fields):
        duplicates = [c for c in fields if fields.count(c) > 1]
        raise DuplicateLabelsException("Columns/index are not unique after renaming! "
                                       f"Duplicates found: {duplicates}")

    # compare columns (and index)
    before = df.reset_index().columns.tolist()
    after = new_df.reset_index().columns.tolist()
    for i, j in zip(before, after):
        if (isinstance(i, str) and isinstance(j, str) and i != j):
            log(f'Renamed column/index "{i}" to "{j}s"')
    return new_df


# ## Function to adjust the size of chunks to upsert
# (depending on a DataFrame's shape and what a database allows for SQL parameters)

def adjust_chunksize(con: Connectable, df: pd.DataFrame, chunksize: int) -> int:
    """
    Checks if given `chunksize` is appropriate for upserting rows in given database using
    given DataFrame.
    The `chunksize` represents the number of rows you would like to upsert in a chunk when
    using the `pangres.upsert` function.

    Some databases have limitations on the number of SQL parameters and we need one parameter
    per value for upserting data.
    E.g. a DataFrame of 5 **columns+index levels** and 4 rows will require 5*4=20 SQL parameters.

    This function will check the database type (e.g. SQlite) and the number of **columns+index levels**
    to determine if the `chunksize` does not exceed limits and propose a lower one if it does.
    Otherwise the same `chunksize` that you gave as input is returned.

    This function currently takes into account max parameters limitations for the following cases:
    * sqlite (32766 max for version >= 3.22.0 otherwise 999)
    * asyncpg (32767 max)

    If you know about more parameter limitations relevant for this library (PostgreSQL, MySQL, SQlite
    or other databases I have not tested with this library that you managed to have working),
    please contact me.

    Parameters
    ----------
    con
        sqlalchemy Engine or Connection
    df
        DataFrame you would wish to upsert
    chunksize
        Size of chunks you would wish to use for upserting (represents the number of rows
        in each chunk)

    Raises
    ------
    TooManyColumnsForUpsertException
        When a DataFrame has more columns+index levels than the maximum number of allowed SQL variables
        for a SQL query for given database.
        In such a case even inserting row by row would not be possible because we would already
        have too many variables.
        For more information you can for instance google "SQLITE_MAX_VARIABLE_NUMBER".

    Examples
    --------
    >>> from sqlalchemy import create_engine
    >>>
    >>> # config (this assumes you have SQlite version >= 3.22.0)
    >>> engine = create_engine("sqlite://")
    >>>
    >>> # some df we want to upsert
    >>> df = pd.DataFrame({'name':['Albert']}).rename_axis(index='profileid')
    >>> print(df.to_markdown())
    |   profileid | name   |
    |------------:|:-------|
    |           0 | Albert |

    {{python}}
    >>> # adjust chunksize: 100,000 is too big of a chunksize in general for given database
    >>> # SQlite only allows 32766 parameters (values) at once maximum in a query
    >>> # since we have two columns (technically 1 column + 1 index level)
    >>> # we can only upsert in chunks of FLOOR(32766/2) rows maximum which is 16383
    >>> adjust_chunksize(con=engine, df=df, chunksize=100_000)
    16383
    """
    validate_chunksize_param(chunksize=chunksize)

    # get maximum number of parameters depending on the database
    dialect = con.dialect.dialect_description  # type: ignore  # dialect attribute does exist
    if 'sqlite' in dialect:
        maximum = 32766 if _sqlite_gt3_32_0() else 999
    elif 'asyncpg' in dialect:
        maximum = 32767
    else:
        maximum = None

    # simple case we can solve early
    if maximum is None:
        return chunksize

    # adjust chunksize
    new_chunksize = floor(maximum / (len(df.columns) + df.index.nlevels))
    if new_chunksize < 1:
        raise TooManyColumnsForUpsertException('The df has more columns+index levels than the maxmimum '
                                               'number of allowed parameters '
                                               'for given database for a query (we could not even upsert row by row).')
    if chunksize > new_chunksize:
        log(f'Reduced chunksize from {chunksize} to {new_chunksize} due '
            f'to max variable restriction of given dialect (max {maximum} for dialect {dialect}).',
            level=logging.INFO)
        chunksize = new_chunksize
    return chunksize



#!/usr/bin/env python
# coding: utf-8
# +
import datetime
import logging
import numpy as np
import pandas as pd
import pytest
from sqlalchemy import create_engine, VARCHAR
from sqlalchemy.sql.elements import Null as SqlaNull

# local imports
from pangres.examples import _TestsExampleTable
from pangres.exceptions import (DuplicateLabelsException,
                                DuplicateValuesInIndexException,
                                HasNoSchemaSystemException,
                                MissingIndexLevelInSqlException,
                                UnnamedIndexLevelsException)
from pangres.engine import PandasSpecialEngine
from pangres.helpers import _sqla_gt14
from pangres.tests.conftest import (adrop_schema, adrop_table_between_tests,
                                    commit, create_sync_or_async_engine,
                                    drop_table_between_tests, drop_schema,
                                    schema_for_testing_creation,
                                    sync_or_async_test, TableNames)


# -

# # Sync and async variants for tests
#
# (`run_test_foo`|`run_test_foo_async`) -> `test_foo`

# ## Schema creation
#
# We do not create any table here

# +
def run_test_schema_creation(engine, schema):
    # overwrite schema
    schema = schema_for_testing_creation

    # schema may already exist before testing
    if 'postgres' in engine.dialect.dialect_description:
        drop_schema(engine=engine, schema=schema)

    # then try to create a schema from a PandasSpecialEngine instance
    dummy_df = pd.DataFrame(index=pd.Index(data=[0], name='id'))
    with engine.connect() as connection:
        try:
            pse = PandasSpecialEngine(connection=connection, schema=schema,
                                      table_name=TableNames.NO_TABLE, df=dummy_df)
            # this should raise HasNoSchemaSystemException
            # if we are not on a postgres engine
            assert not pse.schema_exists()
            pse.create_schema_if_not_exists()
            commit(connection)
            assert pse.schema_exists()
        except Exception as e:
            if pse._db_type == 'postgres' or not isinstance(e, HasNoSchemaSystemException):
                raise e  # pragma: no cover
        finally:
            if pse._db_type == 'postgres':
                drop_schema(engine=engine, schema=schema)


async def run_test_schema_creation_async(engine, schema):
    # overwrite schema
    schema = schema_for_testing_creation

    # schema may already exist before testing
    if 'postgres' in engine.dialect.dialect_description:
        await adrop_schema(engine=engine, schema=schema)

    # then try to create a schema from a PandasSpecialEngine instance
    dummy_df = pd.DataFrame(index=pd.Index(data=[0], name='id'))
    async with engine.connect() as connection:
        try:
            pse = PandasSpecialEngine(connection=connection, schema=schema,
                                      table_name=TableNames.NO_TABLE, df=dummy_df)
            # this should raise HasNoSchemaSystemException
            # if we are not on a postgres engine
            assert not await pse.aschema_exists()
            await pse.acreate_schema_if_not_exists()
            await connection.commit()
            assert await pse.aschema_exists()
        except Exception as e:
            if pse._db_type == 'postgres' or not isinstance(e, HasNoSchemaSystemException):
                raise e  # pragma: no cover
        finally:
            if pse._db_type == 'postgres':
                await adrop_schema(engine=engine, schema=schema)


# -

# ## Table creation

# +
@drop_table_between_tests(table_name=TableNames.TABLE_CREATION)
def run_test_table_creation(engine, schema):
    df = _TestsExampleTable.create_example_df(nb_rows=10)
    with engine.connect() as connection:
        pse = PandasSpecialEngine(connection=connection, schema=schema,
                                  table_name=TableNames.TABLE_CREATION, df=df)
        assert not pse.table_exists()
        pse.create_table_if_not_exists()
        commit(connection)
        assert pse.table_exists()


@adrop_table_between_tests(table_name=TableNames.TABLE_CREATION)
async def run_test_table_creation_async(engine, schema):
    df = _TestsExampleTable.create_example_df(nb_rows=10)
    async with engine.connect() as connection:
        pse = PandasSpecialEngine(connection=connection, schema=schema,
                                  table_name=TableNames.TABLE_CREATION, df=df)
        assert not await pse.atable_exists()
        await pse.acreate_table_if_not_exists()
        await connection.commit()
        assert await pse.atable_exists()


# -

# ## Adding new columns

# +
@drop_table_between_tests(table_name=TableNames.ADD_NEW_COLUMN)
def run_test_add_new_columns(engine, schema, on_index: bool):
    # store arguments we will use for multiple PandasSpecialEngine instances
    table_name = TableNames.ADD_NEW_COLUMN
    common_kwargs = dict(schema=schema, table_name=table_name)

    # create our example table
    df = _TestsExampleTable.create_example_df(nb_rows=10)
    with engine.connect() as connection:
        pse = PandasSpecialEngine(connection=connection, df=df, **common_kwargs)
        pse.create_table_if_not_exists()
        commit(connection)
        assert pse.table_exists()

    # we need to recreate an instance of PandasSpecialEngine
    # so that a new table model with the new columns is created then add columns
    with engine.connect() as connection:
        # add a new index level or new columns (no JSON ones,
        # it's not supported by sqlalchemy compilers :( )
        if on_index:
            df['new_index_col'] = 'foo'
            df.set_index('new_index_col', append=True, inplace=True)
        else:
            df = df.assign(new_text_col='test',
                           new_int_col=0,
                           new_float_col=1.1,
                           new_bool_col=False,
                           new_dt_col=pd.Timestamp('2020-01-01'),
                           # create this col for later
                           empty_col=None)

        # recreate PandasSpecialEngine
        pse = PandasSpecialEngine(connection=connection, df=df, **common_kwargs)

        # check if we get an error when trying to add an index level
        if on_index:
            with pytest.raises(MissingIndexLevelInSqlException) as exc_info:
                pse.add_new_columns()
            assert 'Cannot add' in str(exc_info.value)
            return
        else:
            pse.add_new_columns()
            commit(connection)

    # check if the columns were correctly added
    assert not on_index
    with engine.connect() as connection:
        df_db = _TestsExampleTable.read_from_db(engine=engine, schema=schema, table_name=table_name)
        assert set(df.columns) == set(df_db.columns)


@adrop_table_between_tests(table_name=TableNames.ADD_NEW_COLUMN)
async def run_test_add_new_columns_async(engine, schema, on_index: bool):
    # store arguments we will use for multiple PandasSpecialEngine instances
    table_name = TableNames.ADD_NEW_COLUMN
    common_kwargs = dict(schema=schema, table_name=table_name)

    # create our example table
    df = _TestsExampleTable.create_example_df(nb_rows=10)
    async with engine.connect() as connection:
        pse = PandasSpecialEngine(connection=connection, df=df, **common_kwargs)
        await pse.acreate_table_if_not_exists()
        await connection.commit()
        assert await pse.atable_exists()

    # we need to recreate an instance of PandasSpecialEngine
    # so that a new table model with the new columns is created then add columns
    async with engine.connect() as connection:
        # add a new index level or new columns (no JSON ones,
        # it's not supported by sqlalchemy compilers :( )
        if on_index:
            df['new_index_col'] = 'foo'
            df.set_index('new_index_col', append=True, inplace=True)
        else:
            df = df.assign(new_text_col='test',
                           new_int_col=0,
                           new_float_col=1.1,
                           new_bool_col=False,
                           new_dt_col=pd.Timestamp('2020-01-01'),
                           # create this col for later
                           empty_col=None)

        # recreate PandasSpecialEngine with the new df
        pse = PandasSpecialEngine(connection=connection, df=df, **common_kwargs)

        # check if we get an error when trying to add an index level
        if on_index:
            with pytest.raises(MissingIndexLevelInSqlException) as exc_info:
                await pse.aadd_new_columns()
            assert 'Cannot add' in str(exc_info.value)
            return
        else:
            await pse.aadd_new_columns()
            await connection.commit()

    # check if the columns were correctly added
    assert not on_index
    async with engine.connect() as connection:
        df_db = await _TestsExampleTable.aread_from_db(engine=engine, schema=schema, table_name=table_name)
        assert set(df.columns) == set(df_db.columns)


# -

# ## Changing data type for empty columns

# +
@drop_table_between_tests(table_name=TableNames.CHANGE_EMPTY_COL_TYPE)
def run_test_change_column_type_if_column_empty(engine, schema, caplog, new_empty_column_value):
    # store arguments we will use for multiple PandasSpecialEngine instances
    table_name = TableNames.CHANGE_EMPTY_COL_TYPE
    common_kwargs = dict(schema=schema, table_name=table_name)
    common_kwargs['dtype'] = {'profileid': VARCHAR(5)} if 'mysql' in engine.dialect.dialect_description else None

    # json like will not work for sqlalchemy < 1.4
    # also skip sqlite as it does not support such alteration
    json_like = isinstance(new_empty_column_value, (dict, list))
    if json_like and not _sqla_gt14():
        pytest.skip('JSON like values will not work for sqlalchemy < 1.4')  # pragma: no cover
    elif 'sqlite' in engine.dialect.dialect_description:
        pytest.skip('such column alteration is not possible with SQlite')

    # create our example table
    df = pd.DataFrame({'profileid': ['foo'], 'empty_col': [None]}).set_index('profileid')
    with engine.connect() as connection:
        pse = PandasSpecialEngine(connection=connection, df=df, **common_kwargs)
        pse.create_table_if_not_exists()
        commit(connection)
        assert pse.table_exists()

    # recreate an instance of PandasSpecialEngine with a new df (so the model gets refreshed)
    # the line below is a "hack" to set any type of element as a column value
    # without pandas trying to broadcast it. This is useful when passing a list or such
    df['empty_col'] = df.index.map(lambda x: new_empty_column_value)
    with engine.connect() as connection:
        pse = PandasSpecialEngine(connection=connection, df=df, **common_kwargs)
        with caplog.at_level(logging.INFO, logger='pangres'):
            pse.adapt_dtype_of_empty_db_columns()
        assert len(caplog.records) == 1
        assert 'Changed type of column empty_col' in caplog.text
    caplog.clear()


@adrop_table_between_tests(table_name=TableNames.CHANGE_EMPTY_COL_TYPE)
async def run_test_change_column_type_if_column_empty_async(engine, schema, caplog, new_empty_column_value):
    # store arguments we will use for multiple PandasSpecialEngine instances
    table_name = TableNames.CHANGE_EMPTY_COL_TYPE
    common_kwargs = dict(schema=schema, table_name=table_name)
    common_kwargs['dtype'] = {'profileid': VARCHAR(5)} if 'mysql' in engine.dialect.dialect_description else None

    # json like will not work for sqlalchemy < 1.4
    # also skip sqlite as it does not support such alteration
    json_like = isinstance(new_empty_column_value, (dict, list))
    if json_like and not _sqla_gt14():
        pytest.skip('JSON like values will not work for sqlalchemy < 1.4')  # pragma: no cover
    elif 'sqlite' in engine.dialect.dialect_description:
        pytest.skip('such column alteration is not possible with SQlite')

    # create our example table
    df = pd.DataFrame({'profileid': ['foo'], 'empty_col': [None]}).set_index('profileid')
    async with engine.connect() as connection:
        pse = PandasSpecialEngine(connection=connection, df=df, **common_kwargs)
        await pse.acreate_table_if_not_exists()
        await connection.commit()
        assert await pse.atable_exists()

    # recreate an instance of PandasSpecialEngine with a new df (so the model gets refreshed)
    # the line below is a "hack" to set any type of element as a column value
    # without pandas trying to broadcast it. This is useful when passing a list or such
    df['empty_col'] = df.index.map(lambda x: new_empty_column_value)
    async with engine.connect() as connection:
        pse = PandasSpecialEngine(connection=connection, df=df, **common_kwargs)
        with caplog.at_level(logging.INFO, logger='pangres'):
            await pse.aadapt_dtype_of_empty_db_columns()
        assert len(caplog.records) == 1
        assert 'Changed type of column empty_col' in caplog.text
    caplog.clear()


# -

# # Actual tests

# ## SQL flavor dependant tests

# +
def test_schema_creation(engine, schema):
    sync_or_async_test(engine=engine, schema=schema,
                       f_async=run_test_schema_creation_async,
                       f_sync=run_test_schema_creation)


def test_table_creation(engine, schema):
    sync_or_async_test(engine=engine, schema=schema,
                       f_async=run_test_table_creation_async,
                       f_sync=run_test_table_creation)


@pytest.mark.parametrize('on_index', [True, False], ids=['in df index', 'not in df index'])
def test_add_new_columns(engine, schema, on_index):
    sync_or_async_test(engine=engine, schema=schema,
                       f_async=run_test_add_new_columns_async,
                       f_sync=run_test_add_new_columns,
                       on_index=on_index)


params_new_value_empty_col = [1, 1.1, pd.Timestamp("2020-01-01", tz='UTC'), {'foo': 'bar'}, ['foo'], True]


@pytest.mark.parametrize(argnames="new_empty_column_value", argvalues=params_new_value_empty_col,
                         ids=[f'new_value={p}' for p in params_new_value_empty_col])
def test_change_column_type_if_column_empty(engine, schema, caplog, new_empty_column_value):
    sync_or_async_test(engine=engine, schema=schema,
                       f_async=run_test_change_column_type_if_column_empty_async,
                       f_sync=run_test_change_column_type_if_column_empty,
                       caplog=caplog, new_empty_column_value=new_empty_column_value)


# -

# ## SQL flavor agnostic tests

# +
def test_values_conversion(_):
    engine = create_engine('sqlite:///')
    row = {'id': 0,
           'pd_interval': pd.Interval(left=0, right=5),
           'nan': np.nan,
           'nat': pd.NaT,
           'none': None,
           'pd_na': getattr(pd, 'NA', None),
           'ts': pd.Timestamp('2021-01-01')}
    df = pd.DataFrame([row]).set_index('id')
    with engine.connect() as connection:
        pse = PandasSpecialEngine(connection=connection, df=df, table_name=TableNames.NO_TABLE)
        values = pse._get_values_to_insert()
    converted_row = values[0]
    assert len(row) == len(converted_row)
    # iterate over the keys of the row and get the index (after conversion we just have a list)
    for ix, k in enumerate(row):
        v = row[k]
        v_converted = converted_row[ix]
        if k == 'id':
            assert v_converted == v
        elif k == 'pd_interval':
            assert isinstance(v_converted, str)
        elif k == 'ts':
            assert isinstance(v_converted, datetime.datetime)
        # we should receive all null likes here
        else:
            assert pd.isna(v)
            assert isinstance(v_converted, SqlaNull)


# dummy connection strings to test our categorization for databases
params_db_type_tests = [('sqlite://', 'sqlite'),
                        ('sqlite+aiosqlite://', 'sqlite'),
                        ('postgresql+psycopg2://username:password@localhost:5432/postgres', 'postgres'),
                        ('postgresql://username:password@localhost:5432/postgres', 'postgres'),
                        ('mysql+pymysql://username:password@localhost:3306/db', 'mysql'),
                        ('mysql+aiomysql://username:password@localhost:3306/db', 'mysql'),
                        ('oracle+cx_oracle://username:password@localhost', 'other')]


@pytest.mark.parametrize("connection_string, expected", params_db_type_tests)
def test_detect_db_type(_, connection_string, expected):
    # there are some engines that we will not be able to create because they are asynchronous
    # and this requires sqlalchemy >= 1.4
    try:
        engine = create_sync_or_async_engine(connection_string)
    except NotImplementedError as e:  # pragma: no cover
        pytest.skip(str(e))
    except ModuleNotFoundError as e:  # pragma: no cover
        raise ModuleNotFoundError('There seems to be a dependency missing for this test. Please install it.') from e
    assert PandasSpecialEngine._detect_db_type(connectable=engine) == expected


def test_repr(_):
    engine = create_engine('sqlite://')
    dummy_df = pd.DataFrame(index=pd.Index(data=[0], name='id'))
    with engine.connect() as connection:
        pse = PandasSpecialEngine(connection=connection, table_name=TableNames.NO_TABLE, df=dummy_df)
        # make sure it is printable without errors
        txt = str(pse)
        print(txt)
        # test some strings we expect to find in the repr
        for s in ('PandasSpecialEngine', 'id ', 'hexid', 'connection',
                  'schema', 'table', 'SQLalchemy table model'):
            assert s in txt


def test_table_attr(_):
    engine = create_engine('sqlite://')
    # generate a somewhat complex table model via the _TestsExampleTable class
    df = _TestsExampleTable.create_example_df(nb_rows=10)
    table_name = TableNames.NO_TABLE
    with engine.connect() as connection:
        pse = PandasSpecialEngine(connection=connection, table_name=table_name, df=df)
        # make sure columns and table name match
        expected_cols = list(df.index.names) + df.columns.tolist()
        assert all((col in pse.table.columns for col in expected_cols))
        assert pse.table.name == table_name


# -

# ### Test errors

# +
@pytest.mark.parametrize("multi_index", [True, False], ids=['multi index', 'single index'])
def test_error_index_level_no_name(_, multi_index):
    engine = create_engine('sqlite://')
    df = pd.DataFrame({'test': [0]})
    if multi_index:
        df.set_index('test', append=True, inplace=True)
    with pytest.raises(UnnamedIndexLevelsException) as excinfo:
        with engine.connect() as connection:
            PandasSpecialEngine(connection=connection, table_name=TableNames.NO_TABLE, df=df)
    assert "All index levels must be named" in str(excinfo.value)


@pytest.mark.parametrize("option", ['index and column collision', 'columns duplicated', 'index duplicated'])
def test_duplicated_names(_, option):
    engine = create_engine('sqlite://')
    df = pd.DataFrame({'test': [0]})
    if option == 'index and column collision':
        df.index.name = 'test'
    elif option == 'columns duplicated':
        df.index.name = 'ix'
        df = df[['test', 'test']]
    elif option == 'index duplicated':
        df = df.set_index(['test', 'test'])
    else:  # pragma: no cover
        raise AssertionError(f'Unexpected value for param `option`: {option}')

    with pytest.raises(DuplicateLabelsException) as excinfo:
        with engine.connect() as connection:
            PandasSpecialEngine(connection=connection, table_name=TableNames.NO_TABLE, df=df)
    assert "Found duplicates across index and columns" in str(excinfo.value)


def test_non_unique_index(_):
    engine = create_engine('sqlite://')
    df = pd.DataFrame(index=pd.Index(data=[0, 0], name='ix'))
    with pytest.raises(DuplicateValuesInIndexException) as excinfo:
        with engine.connect() as connection:
            PandasSpecialEngine(connection=connection, table_name=TableNames.NO_TABLE, df=df)
    assert "The index must be unique" in str(excinfo.value)


@pytest.mark.parametrize("bad_chunksize_value", [0, -1, 1.2])
def test_bad_chunksize(_, bad_chunksize_value):
    engine = create_engine('sqlite://')
    df = pd.DataFrame({'test': [0]})
    df.index.name = 'id'
    with engine.connect() as connection:
        pse = PandasSpecialEngine(connection=connection, table_name=TableNames.NO_TABLE, df=df)
        with pytest.raises(ValueError) as excinfo:
            pse._create_chunks(values=[0], chunksize=bad_chunksize_value)
        assert "integer strictly above 0" in str(excinfo.value)



#!/usr/bin/env python
# coding: utf-8
# +
"""
Configuration and helpers for the tests of pangres with pytest.
"""
import asyncio
import inspect
import pandas as pd
import sqlalchemy as sa
from functools import wraps
from inspect import signature
from sqlalchemy import create_engine, text
from sqlalchemy.sql.compiler import IdentifierPreparer
from typing import Union

# local imports
from pangres.helpers import _sqla_gt14
# -

# # Helpers for other test modules

# ## Async/Sync detection and handling

# +
async_sql_drivers = ('asyncpg', 'aiosqlite', 'aiomysql')


def execute_coroutine_sync(coro):
    """
    Retrieves or creates an event loop and executes given coroutine
    synchronously. The strategy is as follows:

    1. try to find an event loop (it may be an event loop we previously created)
    2. if there is no event loop (this is the case when pytest starts it seems), create one
    3. execute coro
    """
    try:
        loop = asyncio.get_event_loop()
    except RuntimeError:  # pragma: no cover
        loop = asyncio.new_event_loop()
        asyncio.set_event_loop(loop=loop)
    task = asyncio.ensure_future(coro, loop=loop)
    return loop.run_until_complete(task)


def sync_async_exec_switch(func, *args, **kwargs):
    """
    Executes given function with given parameters synchronously regardless
    whether it is a coroutine or not.

    Examples
    --------
    >>> def foo():
    ...     print('test sync')
    >>>
    >>> async def foo_async():
    ...     print('test async')
    >>>
    >>> sync_async_exec_switch(foo)
    test sync
    >>> sync_async_exec_switch(foo_async)
    test async
    """
    if inspect.iscoroutine(func) or inspect.iscoroutinefunction(func):
        return execute_coroutine_sync(func(*args, **kwargs))
    elif callable(func):
        return func(*args, **kwargs)
    else:  # pragma: no cover
        raise TypeError('Expected a coroutine or callable')


def is_async_sqla_obj(obj):
    """
    Returns True if `obj` is an asynchronous sqlalchemy connectable (engine or connection)
    otherwise False.
    """
    # sqla < 1.4 does not support asynchronous connectables
    if not _sqla_gt14():
        return False
    from sqlalchemy.ext.asyncio.engine import AsyncConnection, AsyncEngine
    return isinstance(obj, (AsyncConnection, AsyncEngine))


def create_sync_or_async_engine(conn_string, **kwargs):
    """
    Automatically creates an appropriate engine for given connection string
    (synchronous or asynchronous).

    Examples
    --------
    >>> # sync
    >>> engine = create_sync_or_async_engine("sqlite://")
    >>> # async
    >>> conn_string_async = "postgresql+asyncpg://username:password@localhost:5432/postgres"
    >>> engine = create_sync_or_async_engine(conn_string_async)  # doctest: +SKIP
    """
    # if we see any known async drivers we will create an async engine
    if any(s in conn_string.split('/')[0] for s in async_sql_drivers):
        if not _sqla_gt14():
            raise NotImplementedError('Asynchronous engines require sqlalchemy >= 1.4')

        from sqlalchemy.ext.asyncio import create_async_engine
        return create_async_engine(conn_string)
    # otherwise we will just assume we have to create a sync engine
    else:
        return create_engine(conn_string)


# -

# ## SQL

# ### Misc

# +
def quote_object_name(con, object_name):
    """
    For quoting SQL schema names and SQL table names to protect
    against SQL injection. If you have a better idea for achieving
    that please post an issue.
    """
    return IdentifierPreparer(dialect=con.dialect).quote(object_name)


def get_table_namespace(con, schema, table_name):
    """
    Gets quoted table namespace (`schema.table_name`) to protect against
    SQL injection.
    """
    schema = quote_object_name(con=con, object_name=schema) if schema is not None else None
    table_name = quote_object_name(con=con, object_name=table_name)
    return f'{schema}.{table_name}' if schema is not None else table_name


def table_exists(connection, schema, table_name) -> bool:
    """
    Checks for SQL table existence
    """
    insp = sa.inspect(connection)
    if _sqla_gt14():
        return insp.has_table(schema=schema, table_name=table_name)
    else:
        return table_name in insp.get_table_names(schema=schema)


def commit(connection):
    """
    This is for newer versions of sqlalchemy (2.0 or 1.4 using future engines)
    where connections start with implicit transactions and have a commit
    attribute
    """
    if hasattr(connection, 'commit'):
        connection.commit()


# -

# ### SELECT

# +
def select_table(engine, schema, table_name,
                 error_if_missing=True,
                 **read_sql_kwargs) -> Union[pd.DataFrame, None]:
    """
    Does a simple SELECT * FROM {table} and returns a DataFrame from that.
    Has an option to return None if the table does not exist and
    `error_if_missing` is False.
    """
    ns = get_table_namespace(con=engine, schema=schema, table_name=table_name)
    with engine.connect() as connection:
        # check if the table is present
        if table_exists(connection=connection, schema=schema, table_name=table_name):
            return pd.read_sql(text(f'SELECT * FROM {ns}'), con=connection, **read_sql_kwargs)
        elif error_if_missing:  # pragma: no cover
            raise AssertionError(f'Table {ns} does not exist')
        else:
            return None


async def aselect_table(engine, schema, table_name,
                        error_if_missing=True,
                        **read_sql_kwargs) -> Union[pd.DataFrame, None]:
    """
    Async variant of select table
    """
    index_kwarg = 'index_col'
    for k in read_sql_kwargs:
        if k != index_kwarg:  # pragma: no cover
            raise NotImplementedError(f'Can only handle handle extra kwarg `{index_kwarg}` '
                                      f'for aselect_table. Got {k}')

    ns = get_table_namespace(con=engine, schema=schema, table_name=table_name)
    async with engine.connect() as connection:
        # check if the table is present
        exists_coro = lambda connection: table_exists(connection=connection,
                                                      schema=schema,
                                                      table_name=table_name)
        exists = await connection.run_sync(exists_coro)
        if exists:
            proxy = await connection.execute(text(f'SELECT * FROM {ns}'))
            df = pd.DataFrame(proxy)
            if index_kwarg in read_sql_kwargs:
                df.set_index(read_sql_kwargs[index_kwarg], inplace=True)
            return df
        elif error_if_missing:  # pragma: no cover
            raise AssertionError(f'Table {ns} does not exist')
        else:
            return None


# -

# ### DROP schema

# +
def drop_schema_from_conn(connection, schema):
    """
    Drops given PostgreSQL schema using connection
    """
    schema = quote_object_name(con=connection, object_name=schema)
    connection.execute(text(f'DROP SCHEMA IF EXISTS {schema};'))
    commit(connection)


async def adrop_schema_from_conn(connection, schema):
    """
    Async variant of `drop_schema_from_conn`
    """
    schema = quote_object_name(con=connection, object_name=schema)
    await connection.execute(text(f'DROP SCHEMA IF EXISTS {schema};'))
    await connection.commit()


def drop_schema(engine, schema):
    """
    Drops given PostgreSQL schema using an engine
    """
    with engine.connect() as connection:
        drop_schema_from_conn(connection=connection, schema=schema)


async def adrop_schema(engine, schema):
    """
    Async variant of `drop_schema`
    """
    async with engine.connect() as connection:
        await adrop_schema_from_conn(connection=connection, schema=schema)


# -

# ### DROP table

# +
def drop_table_from_conn(connection, schema, table_name):
    """
    Drops given SQL table using connection
    """
    ns = get_table_namespace(con=connection, schema=schema, table_name=table_name)
    # instead of using `DROP TABLE IF EXISTS` which produces warnings
    # in MySQL when the table does exist, we will just check if it does exist
    # and using a `DROP TABLE` query accordingly
    if not table_exists(connection=connection, schema=schema, table_name=table_name):
        return
    # make sure to commit this!
    connection.execute(text(f'DROP TABLE {ns};'))
    commit(connection)


def drop_table(engine, schema, table_name):
    """
    Drops given SQL table using engine
    """
    with engine.connect() as connection:
        drop_table_from_conn(connection=connection, schema=schema, table_name=table_name)


async def adrop_table(engine, schema, table_name):
    """
    Async variant of `drop_table`
    """
    drop_coro = lambda connection: drop_table_from_conn(connection=connection, schema=schema,
                                                        table_name=table_name)
    async with engine.connect() as connection:
        await connection.run_sync(drop_coro)


def _get_function_param_value(function, param_name, args, kwargs):
    """
    Helper for function `drop_table_between_tests` to
    retrieve the value of a specific parameter
    from *args and **kwargs like when we wrap a function

    Examples
    --------
    >>> def dummy_decorator(function):
    ...     def wrapper(*args, **kwargs):
    ...         for param_name in ('a', 'b', 'c'):  # see function `foo` below
    ...             value = _get_function_param_value(function=function, param_name=param_name,
    ...                                               args=args, kwargs=kwargs)
    ...             print(f'{param_name}={value}')
    ...     return wrapper
    >>>
    >>> @dummy_decorator
    ... def foo(a, b, c):
    ...     pass
    >>>
    >>> foo(1, 2, c=3)
    a=1
    b=2
    c=3
    """
    sig = signature(function)
    assert param_name in sig.parameters, f'Parameter "{param_name}" not found in function {function.__name__}!'
    param_ix = list(sig.parameters).index(param_name)
    if param_ix <= len(args) - 1:
        return args[param_ix]
    else:
        return kwargs[param_name]


def drop_table_between_tests(table_name, drop_before_test=True, drop_after_test=True):
    """
    Decorator for wrapping our tests functions. Expects the tests function
    to have the arguments "engine" and "schema" and will retrieve their values
    for dropping the table named `table_name`.

    Parameters
    ----------
    drop_before_test
        Whether to drop the table before executing the test
    drop_after_test
        Whether to drop the table after executing the test
    """
    assert drop_before_test or drop_after_test, 'One of drop_before_test or drop_after_test must be True'

    def sub_decorator(function):
        @wraps(function)
        def wrapper(*args, **kwargs):
            # get engine and schema
            engine = _get_function_param_value(function=function, param_name='engine', args=args, kwargs=kwargs)
            schema = _get_function_param_value(function=function, param_name='schema', args=args, kwargs=kwargs)
            # before test
            if drop_before_test:
                drop_table(engine=engine, schema=schema, table_name=table_name)
            # test
            result = function(*args, **kwargs)
            # after test
            if drop_after_test:
                drop_table(engine=engine, schema=schema, table_name=table_name)
            return result
        return wrapper
    return sub_decorator


def adrop_table_between_tests(table_name, drop_before_test=True, drop_after_test=True):
    """
    Async variant of `drop_table_between_tests`
    """
    assert drop_before_test or drop_after_test, 'One of drop_before_test or drop_after_test must be True'

    def sub_decorator(function):
        @wraps(function)
        async def wrapper(*args, **kwargs):
            # get engine and schema
            engine = _get_function_param_value(function=function, param_name='engine', args=args, kwargs=kwargs)
            schema = _get_function_param_value(function=function, param_name='schema', args=args, kwargs=kwargs)
            # before test
            if drop_before_test:
                await adrop_table(engine=engine, schema=schema, table_name=table_name)
            # test
            result = await function(*args, **kwargs)
            # after test
            if drop_after_test:
                await adrop_table(engine=engine, schema=schema, table_name=table_name)
            return result
        return wrapper
    return sub_decorator


# -

# ## Variables for tests

# +
class TableNames:
    """
    Names of tables used for tests
    """
    ADD_NEW_COLUMN = 'test_add_new_column'
    BAD_COLUMN_NAMES = 'test_bad_column_names'
    BAD_COLUMN_NAMES_PG = 'test_bad_column_names_pg'
    BAD_TEXT = 'test_bad_text'
    BENCHMARK_INSERT = 'test_speed_insert'
    BENCHMARK_UPSERT = 'test_speed_upsert'
    CHANGE_EMPTY_COL_TYPE = 'test_change_empty_col_type'
    COLUMN_NAMED_VALUES = 'test_column_named_values'
    COMMIT_AS_YOU_GO = 'test_commit_as_you_go'
    COMMIT_OR_ROLLBACK_TRANS = 'test_commit_or_rollback_trans'
    CREATE_SCHEMA_NONE = 'test_create_schema_none'
    CREATE_SCHEMA_NOT_NONE = 'test_create_schema_not_none'
    END_TO_END = 'test_end_to_end'
    INDEX_ONLY_INSERT = 'test_index_only_insert'
    INDEX_WITH_NULL = 'test_index_with_null'
    MULTIINDEX = 'test_multiindex'
    PK_MYSQL = 'test_pk_mysql'  # for checking if autoincrement is disabled in MySQL
    REUSE_CONNECTION = 'test_reuse_connection'
    TABLE_CREATION = 'test_table_creation'
    UNIQUE_KEY = 'test_unique_key'
    VARIOUS_CHUNKSIZES = 'test_chunksize'
    WITH_YIELD = 'test_with_yield'
    WITH_YIELD_EMPTY = 'test_with_yield_empty'
    # case when we need a table name but we are not going to use it for
    # creating/updating a table (if you see it in the DB, something went wrong)
    NO_TABLE = 'test_no_table'


# name of the PostgreSQL schema used for testing schema creation
schema_for_testing_creation = 'pangres_create_schema_test'


# -

# ## Function for running tests that have sync and async variants

def sync_or_async_test(engine, schema, f_async, f_sync, **kwargs):
    """
    Will execute the synchronous (Parameter `f_sync`)
    or asynchronous (Parameter `f_async`) test function
    depending on given `engine`.

    We will indeed have many tests taking the parameters
    `engine` and `schema` and having an async and a sync
    variant.

    In order for pytest not to trigger functions `f_async` and
    `f_sync` we will use a prefix `run_` for our async and sync
    variants in the test modules.

    This stategy has the following advantages:
    * we can use pytest parameterize (pytest-asyncio does not seem to support it)
    * we can also use the same test function for sync and async engines
    * in rare cases the tests function for async engines will actually be synchronous
      (e.g. for benchmarks) but this function will still work

    Examples
    --------
    >>> from sqlalchemy import create_engine
    >>>
    >>>
    >>> def run_test_connect(engine, schema):
    ...     with engine.connect():
    ...         pass
    >>>
    >>> async def run_test_connect_async(engine, schema):
    ...     with engine.connect():
    ...         pass
    >>>
    >>> def test_connect(engine, schema):
    ...     sync_or_async_test(engine=engine, schema=schema,
    ...                        f_async=run_test_connect_async,
    ...                        f_sync=run_test_connect)
    >>>
    >>> # sync
    >>> from sqlalchemy import create_engine
    >>> engine = create_engine('sqlite://')
    >>> schema = None
    >>> test_connect(engine=engine, schema=schema)
    >>>
    >>> # async
    >>> from sqlalchemy.ext.asyncio import create_async_engine  # doctest: +SKIP
    >>> engine = create_async_engine('sqlite+aiosqlite://')  # doctest: +SKIP
    >>> test_connect(engine=engine, schema=schema)  # doctest: +SKIP
    """
    test_func = f_async if is_async_sqla_obj(engine) else f_sync
    sync_async_exec_switch(func=test_func, engine=engine, schema=schema, **kwargs)


# # Tests generation
#
# Note: I thought about using `pytest_session_start` for dropping tables instead of using decorators
# to drop tables between tests but this would be problematic if the user wanted to use the same
# database for sync and async tests.

# +
def pytest_addoption(parser):
    parser.addoption('--sqlite_conn', action="store", type=str, default=None)
    parser.addoption('--async_sqlite_conn', action="store", type=str, default=None)
    parser.addoption('--pg_conn', action="store", type=str, default=None)
    parser.addoption('--async_pg_conn', action="store", type=str, default=None)
    parser.addoption('--mysql_conn', action="store", type=str, default=None)
    parser.addoption('--async_mysql_conn', action="store", type=str, default=None)
    parser.addoption('--pg_schema', action='store', type=str, default=None)


def pytest_generate_tests(metafunc):
    # this is called for every test
    # only when we see the parameters "engine" and "schema" in a function
    # will we repeat a given test (at least once for each engine)
    func_params = signature(metafunc.function).parameters
    if not ('engine' in func_params and 'schema' in func_params):
        # I could not find any other way than to add a dummy
        # for executing a test only once (parameterize needs arguments)
        metafunc.parametrize('_', [''], scope='module')
        return

    # tests that we need to repeat for each engine + options (e.g. future)
    conn_strings = {'sqlite': metafunc.config.option.sqlite_conn,
                    'async_sqlite_conn': metafunc.config.option.async_sqlite_conn,
                    'pg': metafunc.config.option.pg_conn,
                    'asyncpg': metafunc.config.option.async_pg_conn,
                    'mysql': metafunc.config.option.mysql_conn,
                    'async_mysql_conn': metafunc.config.option.async_mysql_conn}
    if not any(v is not None for v in conn_strings.values()):  # pragma: no cover
        raise ValueError('You must provide at least one connection string (e.g. argument --sqlite_conn)!')

    # prepare parameters for metafunc.parametrize
    engines, schemas, ids = [], [], []
    for db_type, conn_string in conn_strings.items():

        # cases where we skip tests generation
        if conn_string is None:
            continue

        # get engine and schema
        schema = metafunc.config.option.pg_schema if db_type in ('pg', 'asyncpg') else None
        engine = create_sync_or_async_engine(conn_string)

        # generate tests
        schemas.append(schema)
        engines.append(engine)
        schema_id = '' if schema is None else f'_schema:{schema}'
        ids.append(f'{engine.url.drivername}{schema_id}')

        # for sqlalchemy 1.4+ use future=True to try the future sqlalchemy 2.0
        # do not do this for async engines which already implement 2.0 functionalities
        if _sqla_gt14() and not is_async_sqla_obj(engine):
            future_engine = create_engine(conn_string, future=True)
            schemas.append(schema)
            engines.append(future_engine)
            ids.append(f'{engine.url.drivername}{schema_id}_future')

    # generate tests
    assert len(engines) == len(schemas) == len(ids)
    metafunc.parametrize("engine, schema", list(zip(engines, schemas)), ids=ids, scope='function')



#!/usr/bin/env python
# coding: utf-8
# +
import pytest
from pandas import __version__ as pandas_version
from sqlalchemy import __version__ as sqla_version
from typing import Any, Dict
# local imports
from pangres import aupsert, upsert
from pangres.helpers import _version_equal_or_greater_than
from pangres.examples import _TestsExampleTable
from pangres.tests.conftest import (adrop_table, drop_table_between_tests, drop_table,
                                    execute_coroutine_sync, sync_or_async_test, TableNames)
from pangres.utils import adjust_chunksize


# -


# # Helpers

def skip_if_sqlalchemy_pandas_conflict():
    """
    pandas >= 1.4.0 requires sqlalchemy >= 1.4.0. We will
    skip tests where pandas is required if these conditions
    are not met
    """
    # pandas >= 1.4.0 requires sqlalchemy >= 1.4.0
    pd_1_4_0 = _version_equal_or_greater_than(pandas_version, '1.4.0')
    sqla_1_4_0 = _version_equal_or_greater_than(sqla_version, '1.4.0')
    if pd_1_4_0 and not sqla_1_4_0:
        pytest.skip('pandas >= 1.4.0 requires sqlalchemy >= 1.4.0 '
                    f'(installed: sqlalchemy v{sqla_version}, pandas v{pandas_version})')


# # Sync and async variants for tests
#
# Note that we cannot use the same strategy ((`run_test_foo`|`run_test_foo_async`) -> `test_foo`) as usual here
# because `pytest-benchmark` which is used inside of both sync and async tests must run sync.
#
# If we were to use the event loop for the async test then we would have to create nested event loops for running
# asynchronous benchmark subfunctions (we cannot use the already running loop).
#
# This is possible with `nest_asyncio` but that just complicates things.
#
# What we will then do is use synchronous functions for testing asynchronous engines. And we will make
# synchronous versions of the benchmark subfunctions.

# iterations has to be 1 when there is a setup
pytest_params: Dict[str, Any] = dict(argnames='nb_rows, rounds, iterations', argvalues=[[10, 5, 1], [1_000, 1, 1]],
                                     ids=['many_little_inserts', 'big_insert'])

# ## "Normal" insert speed
#
# Case where the table does not exist yet so there is no pk to compare.


# +
@drop_table_between_tests(table_name=TableNames.BENCHMARK_INSERT)
def run_test_create_and_insert_speed(engine, schema, benchmark, library, nb_rows, rounds, iterations):
    # get a df (we don't test JSON as this is problematic with pandas)
    df = _TestsExampleTable.create_example_df(nb_rows=nb_rows).drop(columns=['favorite_colors'])

    # prepare setup function
    table_name = TableNames.BENCHMARK_INSERT

    def setup():
        drop_table(engine=engine, schema=schema, table_name=table_name)

    # prepare func for benchmark
    chunksize = adjust_chunksize(con=engine, df=df, chunksize=nb_rows)
    if library == 'pangres':
        def benchmark_func():  # pragma: no cover
            upsert(con=engine, df=df, schema=schema, chunksize=chunksize, table_name=table_name,
                   if_row_exists='update')

    elif library == 'pandas':
        skip_if_sqlalchemy_pandas_conflict()

        def benchmark_func():  # pragma: no cover
            # create table
            df.to_sql(con=engine, schema=schema, name=table_name, method='multi',
                      chunksize=chunksize)

    # benchmark
    try:
        benchmark.pedantic(benchmark_func, setup=setup, rounds=rounds, iterations=iterations)
    except NotImplementedError as e:  # pragma: no cover
        if 'not implemented for SQLAlchemy 2' in str(e):
            pytest.skip('in Python 3.6 there is some kind of problem with engines created with '
                        '`future=True` flag and pandas')


# IMPORTANT1: not "async def"!
# IMPORTANT2: no decorator for dropping table!
def run_test_create_and_insert_speed_async(engine, schema, benchmark, library, nb_rows, rounds, iterations):
    if library == 'pandas':
        pytest.skip('async engines will not work with pandas')

    # get a df (we don't test JSON as this is problematic with pandas)
    df = _TestsExampleTable.create_example_df(nb_rows=nb_rows).drop(columns=['favorite_colors'])

    # prepare setup function (make it synchronous even if the engine is asynchronous)
    table_name = TableNames.BENCHMARK_INSERT

    def drop_table_sync():
        execute_coroutine_sync(adrop_table(engine=engine, schema=schema, table_name=table_name))

    setup = drop_table_sync

    # prepare func for benchmark and its synchronous alternative
    chunksize = adjust_chunksize(con=engine, df=df, chunksize=nb_rows)
    if library == 'pangres':

        async def abenchmark_func():  # pragma: no cover
            await aupsert(con=engine, df=df, schema=schema, chunksize=chunksize,
                          table_name=table_name, if_row_exists='update')

        def benchmark_func():  # pragma: no cover
            execute_coroutine_sync(abenchmark_func())

    else:  # pragma: no cover
        raise AssertionError('This test can only work for the `pangres` library')

    # benchmark
    try:
        benchmark.pedantic(benchmark_func, setup=setup, rounds=rounds, iterations=iterations)
    except NotImplementedError as e:  # pragma: no cover
        if 'not implemented for SQLAlchemy 2' in str(e):
            pytest.skip('in Python 3.6 there is some kind of problem with engines created with '
                        '`future=True` flag and pandas')
        else:
            raise e
    finally:
        drop_table_sync()


# -

# ## Upsert overwrite speed
#
# This feature is not available in pandas yet

# +
@drop_table_between_tests(table_name=TableNames.BENCHMARK_UPSERT)
def run_test_upsert_speed(engine, schema, benchmark, library, nb_rows, rounds, iterations, if_row_exists):
    assert library == 'pangres'  # in case pandas changes and we forget to update the tests

    # get a df
    df = _TestsExampleTable.create_example_df(nb_rows=nb_rows).drop(columns=['favorite_colors'])

    # prepare setup function
    chunksize = adjust_chunksize(con=engine, df=df, chunksize=nb_rows)
    table_name = TableNames.BENCHMARK_UPSERT
    common_kwargs_upsert = dict(con=engine, df=df, schema=schema, chunksize=chunksize,
                                table_name=table_name, if_row_exists=if_row_exists)

    def setup():
        drop_table(engine=engine, schema=schema, table_name=table_name)
        # create with data so that we can measure the impact of looking up and overwriting or
        # ignoring records based on primary keys
        upsert(**common_kwargs_upsert, create_table=True)

    # prepare func for benchmark
    # insert update/ignore with `create_table=False` to maximise speed
    def benchmark_func():  # pragma: no cover
        upsert(**common_kwargs_upsert, create_table=False)

    # benchmark
    benchmark.pedantic(benchmark_func, setup=setup, rounds=rounds, iterations=iterations)


# IMPORTANT1: not "async def"!
# IMPORTANT2: no decorator for dropping table!
def run_test_upsert_speed_async(engine, schema, benchmark, library, nb_rows, rounds, iterations, if_row_exists):
    assert library == 'pangres'  # in case pandas changes and we forget to update the tests

    # get a df
    df = _TestsExampleTable.create_example_df(nb_rows=nb_rows).drop(columns=['favorite_colors'])

    # prepare setup function (make it synchronous even if the engine is asynchronous)
    chunksize = adjust_chunksize(con=engine, df=df, chunksize=nb_rows)
    table_name = TableNames.BENCHMARK_UPSERT
    common_kwargs_upsert = dict(con=engine, df=df, schema=schema, chunksize=chunksize,
                                table_name=table_name, if_row_exists=if_row_exists)

    def drop_table_sync():
        execute_coroutine_sync(adrop_table(engine=engine, schema=schema, table_name=table_name))

    def setup():
        drop_table_sync()
        # create with data so that we can measure the impact of looking up and overwriting or
        # ignoring records based on primary keys
        execute_coroutine_sync(aupsert(**common_kwargs_upsert, create_table=True))

    # prepare func for benchmark
    # insert update/ignore with `create_table=False` to maximise speed
    def benchmark_func():  # pragma: no cover
        execute_coroutine_sync(aupsert(**common_kwargs_upsert, create_table=False))

    # benchmark
    try:
        benchmark.pedantic(benchmark_func, setup=setup, rounds=rounds, iterations=iterations)
    finally:
        drop_table_sync()


# -

# # Actual tests

# +
@pytest.mark.parametrize('library', ['pandas', 'pangres'])
@pytest.mark.parametrize(**pytest_params)
def test_create_and_insert(engine, schema, benchmark, library, nb_rows, rounds, iterations):
    sync_or_async_test(engine=engine, schema=schema,
                       f_async=run_test_create_and_insert_speed_async,
                       f_sync=run_test_create_and_insert_speed,
                       benchmark=benchmark, library=library,
                       nb_rows=nb_rows, rounds=rounds, iterations=iterations)


@pytest.mark.parametrize('library', ['pangres'])
@pytest.mark.parametrize('if_row_exists', ['update', 'ignore'])
@pytest.mark.parametrize(**pytest_params)
def test_upsert_speed(engine, schema, benchmark, library, nb_rows, rounds, iterations, if_row_exists):
    sync_or_async_test(engine=engine, schema=schema,
                       f_async=run_test_upsert_speed_async,
                       f_sync=run_test_upsert_speed,
                       benchmark=benchmark, library=library,
                       nb_rows=nb_rows, rounds=rounds, iterations=iterations,
                       if_row_exists=if_row_exists)



# -*- coding: utf-8 -*-
import pytest
from pangres.helpers import validate_chunksize_param


# # Tests

@pytest.mark.parametrize('value', [-1, 0, 10, 'abc'])
def test_valid_chunksize_values(_, value):
    # 10 is the only valid value here
    if value == 10:
        validate_chunksize_param(value)
        return
    # all the other values should fail
    with pytest.raises((TypeError, ValueError)):
        validate_chunksize_param(value)



# +
"""
This module tests different ways of using pangres by providing a connection
instead of an engine and by using transactions
"""
import pandas as pd
import pytest
from sqlalchemy import text, VARCHAR

# local imports
from pangres import aupsert, upsert
from pangres.helpers import _sqla_gt14
from pangres.transaction import TransactionHandler
from pangres.tests.conftest import (aselect_table, adrop_table_between_tests, commit, drop_table_between_tests,
                                    select_table, sync_or_async_test, sync_async_exec_switch, TableNames)


# -

# # Sync and async variants for tests
#
# (`run_test_foo`|`run_test_foo_async`) -> `test_foo`

# ## Reusing a connection

# +
@drop_table_between_tests(table_name=TableNames.REUSE_CONNECTION)
def run_test_connection_usable_after_upsert(engine, schema):
    df = pd.DataFrame(index=pd.Index([0], name='ix'))
    with engine.connect() as con:
        # do some random upsert operation
        upsert(con=con, df=df, schema=schema,
               table_name=TableNames.REUSE_CONNECTION,
               if_row_exists='update')
        # attempt to reuse the connection
        result = con.execute(text('SELECT 1;')).scalar()
        assert result == 1
        commit(con)


@adrop_table_between_tests(table_name=TableNames.REUSE_CONNECTION)
async def run_test_connection_usable_after_upsert_async(engine, schema):
    df = pd.DataFrame(index=pd.Index([0], name='ix'))
    async with engine.connect() as con:
        # do some random upsert operation
        await aupsert(con=con, df=df, schema=schema,
                      table_name=TableNames.REUSE_CONNECTION,
                      if_row_exists='update')
        # attempt to reuse the connection
        proxy = await con.execute(text('SELECT 1;'))
        result = proxy.scalar()
        assert result == 1
        await con.commit()


# -

# ## Using our own transaction

# +
@drop_table_between_tests(table_name=TableNames.COMMIT_OR_ROLLBACK_TRANS)
def run_test_transaction(engine, schema, trans_op):
    df = pd.DataFrame(index=pd.Index(['foo'], name='ix'))
    table_name = TableNames.COMMIT_OR_ROLLBACK_TRANS
    # common keyword arguments for multiple upsert operations below
    common_kwargs = dict(schema=schema, table_name=table_name,
                         if_row_exists='update', dtype={'ix': VARCHAR(3)})

    with engine.connect() as connection:
        trans = connection.begin()
        try:
            # do some random upsert operation
            upsert(con=connection, df=df, **common_kwargs)
            # do some other operation that requires commit
            upsert(con=connection, df=df.rename(index={'foo': 'bar'}), **common_kwargs)
            getattr(trans, trans_op)()  # commit or rollback
        finally:
            trans.close()

    # if trans_op=='commit': make sure we have "bar" and "foo" in the index
    # elif trans_op=='rollback': make sure we don't have any data
    # or that the table was not even created (what is rolled back
    # depends on the database type and other factors)
    if trans_op == 'commit':
        df_db = select_table(engine=engine, schema=schema, table_name=table_name, index_col='ix')
        # this simple check will be enough since we do not have values, only an index
        assert df_db.index.sort_values().tolist() == ['bar', 'foo']
        assert df_db.index.name == 'ix'

    elif trans_op == 'rollback':
        df_db = select_table(engine=engine, schema=schema, table_name=table_name, error_if_missing=False)
        # no table or an empty table
        assert df_db is None or len(df_db) == 0


@adrop_table_between_tests(table_name=TableNames.COMMIT_OR_ROLLBACK_TRANS)
async def run_test_transaction_async(engine, schema, trans_op):
    df = pd.DataFrame(index=pd.Index(['foo'], name='ix'))
    table_name = TableNames.COMMIT_OR_ROLLBACK_TRANS
    # common keyword arguments for multiple upsert operations below
    common_kwargs = dict(schema=schema, table_name=table_name,
                         if_row_exists='update', dtype={'ix': VARCHAR(3)})

    async with engine.connect() as connection:
        trans = await connection.begin()
        try:
            # do some random upsert operation
            await aupsert(con=connection, df=df, **common_kwargs)
            # do some other operation that requires commit
            await aupsert(con=connection, df=df.rename(index={'foo': 'bar'}), **common_kwargs)
            coro = getattr(trans, trans_op)  # commit or rollback
            await coro()
        finally:
            await trans.close()

    # if trans_op=='commit': make sure we have "bar" and "foo" in the index
    # elif trans_op=='rollback': make sure we don't have any data
    # or that the table was not even created (what is rolled back
    # depends on the database type and other factors)
    if trans_op == 'commit':
        df_db = await aselect_table(engine=engine, schema=schema, table_name=table_name, index_col='ix')
        # this simple check will be enough since we do not have values, only an index
        assert df_db.index.sort_values().tolist() == ['bar', 'foo']
        assert df_db.index.name == 'ix'

    elif trans_op == 'rollback':
        df_db = await aselect_table(engine=engine, schema=schema, table_name=table_name, error_if_missing=False)
        # no table or an empty table
        assert df_db is None or len(df_db) == 0


# -

# ## Commit-as-you-go or ROLLBACK

# +
@drop_table_between_tests(table_name=TableNames.COMMIT_AS_YOU_GO)
def run_test_commit_as_you_go(engine, schema):
    df = pd.DataFrame(index=pd.Index(['foo'], name='ix'))
    table_name = TableNames.COMMIT_AS_YOU_GO
    # common keyword arguments for multiple upsert operations below
    common_kwargs = dict(schema=schema, table_name=table_name,
                         if_row_exists='update', dtype={'ix': VARCHAR(3)})

    with engine.connect() as connection:
        # skip for sqlalchemy < 2.0 or when future=True flag is not passed
        # during engine creation (commit-as-you-go is a new feature)
        # when this is the case there is no attribute commit or rollback for
        # the connection
        if not hasattr(connection, 'commit'):  # pragma: no cover
            pytest.skip('test not possible because there is no attribute "commit" (most likely sqlalchemy < 2)')

        # do some random upsert operation and commit
        upsert(con=connection, df=df, **common_kwargs)
        connection.commit()

        # do some other operation that requires commit and then rollback
        upsert(con=connection, df=df.rename(index={'foo': 'bar'}), **common_kwargs)
        connection.rollback()

    # the table in the db should be equal to the initial df as the second
    # operation was rolled back
    df_db = select_table(engine=engine, schema=schema, table_name=table_name, index_col='ix')
    # this simple check will be enough since we do not have values, only an index
    assert df_db.index.sort_values().tolist() == df.index.sort_values().tolist()
    assert df_db.index.name == 'ix'


@adrop_table_between_tests(table_name=TableNames.COMMIT_AS_YOU_GO)
async def run_test_commit_as_you_go_async(engine, schema):
    df = pd.DataFrame(index=pd.Index(['foo'], name='ix'))
    table_name = TableNames.COMMIT_AS_YOU_GO
    # common keyword arguments for multiple upsert operations below
    common_kwargs = dict(schema=schema, table_name=table_name,
                         if_row_exists='update', dtype={'ix': VARCHAR(3)})

    async with engine.connect() as connection:
        # do some random upsert operation and commit
        await aupsert(con=connection, df=df, **common_kwargs)
        await connection.commit()

        # do some other operation that requires commit and then rollback
        await aupsert(con=connection, df=df.rename(index={'foo': 'bar'}), **common_kwargs)
        await connection.rollback()

    # the table in the db should be equal to the initial df as the second
    # operation was rolled back
    df_db = await aselect_table(engine=engine, schema=schema, table_name=table_name, index_col='ix')
    # this simple check will be enough since we do not have values, only an index
    assert df_db.index.sort_values().tolist() == df.index.sort_values().tolist()
    assert df_db.index.name == 'ix'


# -

# ## Errors

# +
def run_test_non_connectable_transaction_handler(_):
    with pytest.raises(TypeError) as exc_info:
        with TransactionHandler(connectable='abc'):
            pass  # pragma: no cover
    assert 'sqlalchemy connectable' in str(exc_info)


async def run_test_non_connectable_transaction_handler_async(_):
    with pytest.raises(TypeError) as exc_info:
        async with TransactionHandler(connectable='abc'):
            pass  # pragma: no cover
    assert 'sqlalchemy connectable' in str(exc_info)


# -

# # Actual tests

# +
def test_connection_usable_after_upsert(engine, schema):
    sync_or_async_test(engine=engine, schema=schema,
                       f_async=run_test_connection_usable_after_upsert_async,
                       f_sync=run_test_connection_usable_after_upsert)


@pytest.mark.parametrize("trans_op", ['commit', 'rollback'])
def test_transaction(engine, schema, trans_op):
    sync_or_async_test(engine=engine, schema=schema,
                       f_async=run_test_transaction_async,
                       f_sync=run_test_transaction,
                       trans_op=trans_op)


def test_commit_as_you_go(engine, schema):
    sync_or_async_test(engine=engine, schema=schema,
                       f_async=run_test_commit_as_you_go_async,
                       f_sync=run_test_commit_as_you_go)


@pytest.mark.parametrize("async_", [True, False], ids=['async', 'sync'])
def test_non_connectable_transaction_handler(_, async_):
    if async_:
        if not _sqla_gt14():
            pytest.skip('Cannot execute async test with sqlalchemy < 1.4')
        test_func = run_test_non_connectable_transaction_handler_async
    else:
        test_func = run_test_non_connectable_transaction_handler
    sync_async_exec_switch(test_func, _='')



from pangres.core import aupsert, upsert  # noqa: F401
from pangres.utils import adjust_chunksize, fix_psycopg2_bad_cols  # noqa: F401
from pangres.examples import DocsExampleTable  # noqa: F401
from pangres._version import __version__  # noqa: F401
from pangres.exceptions import (BadColumnNamesException, HasNoSchemaSystemException,  # noqa: F401
                                UnnamedIndexLevelsException,  # noqa: F401
                                DuplicateValuesInIndexException, DuplicateLabelsException,  # noqa: F401
                                MissingIndexLevelInSqlException, TooManyColumnsForUpsertException)  # noqa: F401



# +
"""
Tools for handling transactions in pangres
"""
from sqlalchemy.engine import Connectable, Connection, Engine, Transaction
from typing import Union
# local imports
from pangres.helpers import _sqla_gt14
from pangres.pangres_types import AsyncConnection, AsyncTransaction


# -

# # Transaction Handler

class TransactionHandler:
    """
    Context manager class for transactions in pangres.

    Workflow (the numbering below can be found in the comments):

    1) create connection (Engine is passed) or use connection (Connection is passed)
    2) create transaction (Engine is passed)
    3) (commit|rollback) if we created a transaction
    4) close connection if we created the connection from the engine

    We do not handle transactions from users. They will have to commit/rollback themselves.
    We also do not create transactions when a Connection is passed so that commit-as-you-go
    workflows are possible (the operations of pangres will behave like basic SQL executions
    in sqlalchemy).

    Parameters
    ----------
    connectable
        If a sqlalchemy Engine is passed we create a connection,
        otherwise we raise an error if `connectable` is not  a sqlalchemy Connection.
        AsyncConnectable (sqlalchemy version >= 1.4) is also accepted.

    Examples
    --------
    >>> import pandas as pd
    >>> from sqlalchemy import create_engine, text
    >>>
    >>> # config
    >>> engine = create_engine('sqlite:///')
    >>> ddl = text('CREATE TABLE test (id INT);')
    >>> insert = text('INSERT INTO test (id) VALUES (1), (2)')
    >>>
    >>> # execute table creation (ddl) and insert
    >>> with TransactionHandler(connectable=engine) as trans:
    ...     result_ddl = trans.connection.execute(ddl)
    ...     result_insert = trans.connection.execute(insert)
    >>>
    >>> # display the table
    >>> with engine.connect() as con:
    ...     df = pd.read_sql(text('SELECT * FROM test;'), con=con)
    ...     print(df.to_markdown())
    |    |   id |
    |---:|-----:|
    |  0 |    1 |
    |  1 |    2 |
    """

    def __init__(self, connectable: Connectable):
        # set attrs
        self.connectable = connectable
        # we will set this when we enter the context
        self.connection: Union[AsyncConnection, Connection, None] = None
        self.transaction: Union[AsyncTransaction, Transaction, None] = None

    def _close_resources(self):
        # we are nesting try...finally (see __exit__ method) to ensure
        # both resources get closed, and we get all tracebacks (we'll see "... another exception occurred")
        # IDK if this is the best way to do this but my only other idea was raising two errors at once
        # (in case the transaction and/or the connection does not close) which does not seem user-friendly
        try:
            # close transaction if we created one
            if self.transaction is not None:
                self.transaction.close()
        finally:
            # close connection if we created one
            if self.connection is not None and isinstance(self.connectable, Engine):
                self.connection.close()

    def __enter__(self):

        # 1) get(create) connection
        if isinstance(self.connectable, Engine):
            self.connection = self.connectable.connect()
        elif isinstance(self.connectable, Connection):
            self.connection = self.connectable
        else:
            raise TypeError(f'Expected a sqlalchemy connectable object ({Connectable}). '
                            f'Got {type(self.connectable)}')

        # 2) get transaction
        if isinstance(self.connectable, Engine):
            # handle case where for some reason we would not be able to create a transaction
            try:
                self.transaction = self.connection.begin()
            except Exception as e:  # pragma: no cover
                self._close_resources()
                raise e
        return self

    def _rollback_or_commit(self, exception_occurred: bool) -> None:
        # case where we were inside a transaction from the user.
        # The user will have to handle rollback and commit
        if self.transaction is None:
            return

        # case where we created the transaction
        if exception_occurred:
            self.transaction.rollback()
        else:
            self.transaction.commit()

    def __exit__(self, ex_type, value, traceback) -> bool:
        # make sure __enter__ was properly executed
        if self.connection is None:  # pragma: no cover
            raise AssertionError('No active connection. Perhaps the context manager was not properly entered?')

        # combine step 3) rollback or commit and 4) closing resources
        exception_occurred = ex_type is not None
        try:
            self._rollback_or_commit(exception_occurred=exception_occurred)
        finally:
            self._close_resources()
        return not exception_occurred  # will be reraised if False

    # ASYNC VARIANTS of methods above that we will prefix with "a"
    # (note that __aenter__ and __aexit__ are special names for
    # async context managers in Python)
    async def _aclose_resources(self):
        from sqlalchemy.ext.asyncio.engine import AsyncEngine

        try:
            # close transaction if we created one
            if self.transaction is not None:
                await self.transaction.close()
        finally:
            # close connection if we created one
            if self.connection is not None and isinstance(self.connectable, AsyncEngine):
                await self.connection.close()

    async def __aenter__(self) -> 'TransactionHandler':
        # make sure the sqlalchemy version allows for async usage
        # we only need to do this on entry of the context manager
        if not _sqla_gt14():
            raise NotImplementedError('Async usage of sqlalchemy requires version >= 1.4')
        from sqlalchemy.ext.asyncio.engine import AsyncEngine, AsyncConnection, AsyncConnectable

        # similar procedure to __enter__ with different object types
        if isinstance(self.connectable, AsyncEngine):
            self.connection = await self.connectable.connect()
        elif isinstance(self.connectable, AsyncConnection):
            self.connection = self.connectable
        else:
            raise TypeError(f'Expected an async sqlalchemy connectable object ({AsyncConnectable}). '
                            f'Got {type(self.connectable)}')

        if isinstance(self.connectable, AsyncEngine):
            try:
                self.transaction = await self.connection.begin()
            except Exception as e:  # pragma: no cover
                self._close_resources()
                raise e
        return self

    async def _arollback_or_commit(self, exception_occurred: bool) -> None:
        # case where we were inside a transaction from the user.
        # The user will have to handle rollback and commit
        if self.transaction is None:
            return

        # case where we created the transaction
        # use type ignore because this function is only to be used in an async context so for an async transaction
        # (should be AsyncTransaction object, but let's not enforce it with e.g. assert)
        if exception_occurred:
            await self.transaction.rollback()  # type: ignore
        else:
            await self.transaction.commit()  # type: ignore

    async def __aexit__(self, ex_type, exc, tb) -> bool:
        if self.connection is None:  # pragma: no cover
            raise AssertionError('No active connection. Perhaps the context manager was not properly entered?')
        exception_occurred = ex_type is not None
        try:
            await self._arollback_or_commit(exception_occurred=exception_occurred)
        finally:
            await self._aclose_resources()
        return not exception_occurred  # will be reraised if False



# +
"""
Functions for preparing/compiling and executing upsert statements
in different SQL flavors.
"""
from copy import deepcopy
from sqlalchemy.dialects.postgresql import insert as pg_insert, Insert as PgInsert
from sqlalchemy.dialects.mysql.dml import insert as mysql_insert, Insert as MySQLInsert
from sqlalchemy.engine.base import Connection
from sqlalchemy.schema import Table
from sqlalchemy.sql.compiler import SQLCompiler
from typing import Union, Any, Dict

# local imports
from pangres.helpers import _sqla_gt14
from pangres.pangres_types import AsyncConnection


# -

# # Main class `UpsertQuery`

class UpsertQuery:
    """
    Helper for creating and executing UPSERT queries in various SQL flavors

    Examples
    --------
    >>> from pangres.engine import PandasSpecialEngine
    >>> from pangres.examples import DocsExampleTable
    >>> from sqlalchemy import create_engine
    >>>
    >>> engine = create_engine('sqlite://')
    >>> with engine.connect() as connection:
    ...     table = PandasSpecialEngine(connection=connection, df=DocsExampleTable.df_upsert,
    ...                                 table_name='doc_upsert').table
    ...     upq = UpsertQuery(connection=connection, table=table)
    """

    def __init__(self, connection: Union[AsyncConnection, Connection], table: Table) -> None:
        self._verify_connection_like_object(connection=connection)
        self.connection = connection
        self.table = table

    @staticmethod
    def _verify_connection_like_object(connection):
        # handle easy cases first
        if isinstance(connection, Connection):
            return True

        # maybe we are in presence of an asynchronous connection
        is_connection = False  # until proven otherwise
        if _sqla_gt14():
            from sqlalchemy.ext.asyncio.engine import AsyncConnection
            is_connection = isinstance(connection, AsyncConnection)

        # raise if not connection like
        if not is_connection:
            raise TypeError(f'Expected a Connection or AsyncConnection object. Got {type(connection)} instead')

    def _create_pg_query(self, values: list, if_row_exists: str) -> PgInsert:
        insert_stmt = pg_insert(self.table).values(values)
        update_cols = []
        if if_row_exists == 'update':
            update_cols.extend([c.name for c in self.table.columns  # type: ignore
                                if c not in list(self.table.primary_key.columns)])

        # handle case when there is only an index in the DataFrame i.e. no columns to update
        if len(update_cols) == 0:
            if_row_exists = 'ignore'

        # get the appropriate insert statement
        if if_row_exists == 'update':
            return insert_stmt.on_conflict_do_update(index_elements=self.table.primary_key.columns,
                                                     set_={k: insert_stmt.excluded[k] for k in update_cols})
        else:
            return insert_stmt.on_conflict_do_nothing()

    def _create_mysql_query(self, values: list, if_row_exists: str) -> MySQLInsert:
        insert_stmt = mysql_insert(self.table).values(values)
        update_cols: Dict[str, Any] = {}

        if if_row_exists == 'update':
            # thanks to: https://stackoverflow.com/a/58180407/10551772
            # prepare kwargs for on_duplicated_key_update (with kwargs and getattr
            # even "bad" column names will resolve e.g. columns with spaces)
            for col in insert_stmt.table.columns:
                col_name = col.name
                if col_name not in self.table.primary_key:
                    update_cols[col_name] = insert_stmt.inserted[col_name]

        # handle case when there is only an index in the DataFrame i.e. no columns to update
        if len(update_cols) == 0:
            if_row_exists = 'ignore'

        # get the appropriate insert statement
        if if_row_exists == 'update':
            upsert = insert_stmt.on_duplicate_key_update(**update_cols)
        else:
            # thanks to: https://stackoverflow.com/a/50870348/10551772
            upsert = insert_stmt.prefix_with('IGNORE')
        return upsert

    def _create_sqlite_query_gt_sqla_14(self, values: list, if_row_exists: str):
        """
        Creates an upsert sqlite query for sqlalchemy>=1.4
        """
        # the next import is not available in sqlalchemy==1.3
        from sqlalchemy.dialects.sqlite import insert as sqlite_insert
        insert_stmt = sqlite_insert(self.table).values(values)
        update_cols = []
        if if_row_exists == 'update':
            update_cols.extend([c.name for c in self.table.columns  # type: ignore
                                if c not in list(self.table.primary_key.columns)])

        # handle case when there is only an index in the DataFrame i.e. no columns to update
        if len(update_cols) == 0:
            if_row_exists = 'ignore'

        # get the appropriate insert statement
        if if_row_exists == 'update':
            return insert_stmt.on_conflict_do_update(index_elements=self.table.primary_key.columns,
                                                     set_={k: insert_stmt.excluded[k] for k in update_cols})
        else:
            return insert_stmt.on_conflict_do_nothing()

    def _create_sqlite_query_sqla_13(self, values: list, if_row_exists: str) -> SQLCompiler:
        """
        Creates an upsert sqlite query for sqlalchemy==1.3
        """
        def escape_col(col):
            # unbound column from its table
            # otherwise the column would compile as "table.col_name"
            # which we could not use in e.g. SQlite
            unbound_col = deepcopy(col)
            unbound_col.table = None
            return str(unbound_col.compile(dialect=self.connection.dialect))

        # prepare start of upsert (INSERT VALUES (...) ON CONFLICT)
        upsert = SQLCompiler(dialect=self.connection.dialect,
                             statement=self.table.insert().values(values))

        # append on conflict clause
        pk = [escape_col(c) for c in self.table.primary_key]  # type: ignore
        non_pks = [escape_col(c) for c in self.table.columns if c not in list(self.table.primary_key)]  # type: ignore
        ondup = f'ON CONFLICT ({",".join(pk)})'
        # always use "DO NOTHING" if there are no primary keys
        if (not non_pks) or (if_row_exists == 'ignore'):
            ondup_action = 'DO NOTHING'
            upsert.string = ' '.join((upsert.string, ondup, ondup_action))
        elif if_row_exists == 'update':
            ondup_action = 'DO UPDATE SET'
            updates = ', '.join(f'{c}=EXCLUDED.{c}' for c in non_pks)
            upsert.string = ' '.join((upsert.string, ondup, ondup_action, updates))
        return upsert

    def _create_sqlite_query(self, values: list, if_row_exists: str) -> SQLCompiler:
        """
        Creates an upsert sqlite query. Uses different implementations depending
        on the sqlalchemy version.
        See helper methods `_create_sqlite_query_gt_sqla_14` and `_create_sqlite_query_sqla_13`.
        """
        method = self._create_sqlite_query_gt_sqla_14 if _sqla_gt14() else self._create_sqlite_query_sqla_13
        return method(values=values, if_row_exists=if_row_exists)  # type: ignore  # parameters are valid

    def create_query(self, db_type: str, values: list, if_row_exists: str):
        r"""
        Helper for creating UPSERT queries in various SQL flavors

        Parameters
        ----------
        db_type
            One of "postgres", "mysql", "sqlite" or "other"
        values
            The structure for the values must match the table defined
            at instantiation (e.g. same order for the columns)
        if_row_exists : {'ignore', 'update'}

        Examples
        --------
        >>> import pandas as pd
        >>> from pangres.engine import PandasSpecialEngine
        >>> from pangres.examples import DocsExampleTable
        >>> from sqlalchemy import create_engine
        >>>
        >>> # config
        >>> test_values = [['foo', 'foo@test.com', pd.Timestamp('2021-01-01', tz='UTC'), 1.4, True, ['blue']]]
        >>> engine = create_engine('sqlite://')
        >>>
        >>> # helpers
        >>> # we want to have a pretty print for our queries, and we also need to do some homogenization because
        >>> # depending on the version of sqlalchemy we will have slightly different compiled queries
        >>> # (but with the same effect)
        >>> pprint_query = lambda query: print(str(query).replace('ON CONFLICT', '\nON CONFLICT')
        ...                                    .replace('excluded', 'EXCLUDED')
        ...                                    .replace(' = ', '=').replace('SET', '\nSET')
        ...                                    .replace('ON DUPLICATE', '\nON DUPLICATE'))
        >>>
        >>> # create the query
        >>> with engine.connect() as connection:
        ...     table = PandasSpecialEngine(connection=connection, df=DocsExampleTable.df_upsert,
        ...                                 table_name='doc_upsert').table
        ...     upq = UpsertQuery(connection=connection, table=table)
        ...     query = upq.create_query(db_type='sqlite', values=test_values, if_row_exists='update')
        >>>
        >>> # pretty print query
        >>> pprint_query(query)  # doctest: +NORMALIZE_WHITESPACE
        INSERT INTO doc_upsert (ix, email, ts, float, bool, json) VALUES (?, ?, ?, ?, ?, ?)
        ON CONFLICT (ix) DO UPDATE
        SET email=EXCLUDED.email, ts=EXCLUDED.ts, float=EXCLUDED.float, bool=EXCLUDED.bool, json=EXCLUDED.json

        >>> # unsupported databases will raise a NotImplementedError
        >>> try:
        ...     query = upq.create_query(db_type='oracle',values=test_values, if_row_exists='update')
        ... except Exception as e:
        ...     print(e)
        No query creation method for oracle. Expected one of ['postgres', 'mysql', 'sqlite', 'other']
        """
        query_creation_methods = {"postgres": self._create_pg_query,
                                  "mysql": self._create_mysql_query,
                                  "sqlite": self._create_sqlite_query,
                                  "other": self._create_sqlite_query}
        try:
            return query_creation_methods[db_type](values=values, if_row_exists=if_row_exists)
        except KeyError:
            raise NotImplementedError(f'No query creation method for {db_type}. '
                                      f'Expected one of {list(query_creation_methods.keys())}')

    def execute(self, db_type: str, values: list, if_row_exists: str):
        query = self.create_query(db_type=db_type, values=values, if_row_exists=if_row_exists)
        return self.connection.execute(query)

    async def aexecute(self, db_type: str, values: list, if_row_exists: str):
        """
        Async variant of method execute
        """
        query = self.create_query(db_type=db_type, values=values, if_row_exists=if_row_exists)
        return await self.connection.execute(query)  # type: ignore  # this is valid



# # Versions checking

# +
def _py_gt3_10() -> bool:
    """
    Returns True if we are running Python >= 3.10 else return False.
    We need that due to different implementations for checking
    versions of other libraries
    """
    import sys
    py_v = sys.version_info
    # first condition is in case Python 4 comes out :P
    return (py_v.major > 3) or ((py_v.major >= 3) and (py_v.minor >= 10))


def _version_equal_or_greater_than(version_string: str, minimal_version_string: str) -> bool:
    """
    Returns True if a library has a version greater or equal than `minimal_version_string`.
    Otherwise returns False.

    To use this function, give the version string of the library in parameter `version_string`.

    Examples
    --------
    >>> _version_equal_or_greater_than('1.1', '2.0')
    False
    >>> _version_equal_or_greater_than('2.1', '2.0')
    True
    """
    if _py_gt3_10():
        from pkg_resources import parse_version
        v = parse_version(version_string)
        min_v = parse_version(minimal_version_string)
        return (v.major, v.minor, v.micro) >= (min_v.major, min_v.minor, min_v.micro)
    else:
        from distutils.version import LooseVersion
        return LooseVersion(version_string) >= LooseVersion(minimal_version_string)


def _sqla_gt14() -> bool:
    """
    Checks if sqlalchemy.__version__ is at least 1.4.0, when several
    deprecations were made.

    Stolen from pandas.io.sql (we don't import it as it's private
    and has just 2 lines of code).
    """
    import sqlalchemy
    return _version_equal_or_greater_than(version_string=sqlalchemy.__version__,
                                          minimal_version_string='1.4.0')


def _sqla_gt20() -> bool:
    """
    Same as function `_sqla_gt14` for checking if sqlalchemy>=2.0.0
    """
    import sqlalchemy
    return _version_equal_or_greater_than(version_string=sqlalchemy.__version__,
                                          minimal_version_string='2.0.0')


def _sqlite_gt3_32_0() -> bool:
    """
    Checks if the SQLite version is >= than 3.32.0.
    Starting from this version we can use more SQL parameters.
    See https://github.com/ThibTrip/pangres/issues/43
    """
    import sqlite3
    return _version_equal_or_greater_than(version_string=sqlite3.sqlite_version,
                                          minimal_version_string='3.32.0')


# -

# # Parameters checking

def validate_chunksize_param(chunksize: int) -> None:
    if not isinstance(chunksize, int):
        raise TypeError(f'Expected chunksize to be an int. Got {type(chunksize)}')
    if chunksize <= 0:
        raise ValueError('chunksize must be strictly above 0')



#!/usr/bin/env python
# coding: utf-8
# +
"""
Here we test different indices:
1. MultiIndex
2. Indices with NULL

# Note
Other problems with indices such as duplicated values are handled by pangres.helpersPandasSpecialEngine.
"""
import pandas as pd
import pytest
from sqlalchemy import VARCHAR
from sqlalchemy.exc import IntegrityError

# local imports
from pangres import aupsert, upsert
from pangres.tests.conftest import (aselect_table, adrop_table_between_tests,
                                    drop_table_between_tests,
                                    select_table, sync_or_async_test, TableNames)
# -


# # Config

# +
# for creating table with MultiIndex
index_col = ['ix1', 'ix2', 'ix3']
df_multiindex = pd.DataFrame({'ix1': [1, 1],
                              'ix2': ['test', 'test2'],
                              'ix3': [pd.Timestamp('2019-01-01'), pd.Timestamp('2019-01-02')],
                              'foo': [1, 2]}).set_index(index_col)

# for inserting values
df_multiindex2 = pd.DataFrame({'ix1': [2, 2],
                               'ix2': ['test', 'test2'],
                               'ix3': [pd.Timestamp('2019-01-01'), pd.Timestamp('2019-01-02')],
                               'foo': [1, 2]}).set_index(index_col)


# -

# # Sync and async variants for tests
#
# (`run_test_foo`|`run_test_foo_async`) -> `test_foo`

# ## Test MultiIndex

# +
@drop_table_between_tests(table_name=TableNames.MULTIINDEX)
def run_test_create_and_insert_table_multiindex(engine, schema):
    # local helpers
    table_name = TableNames.MULTIINDEX

    def read_from_db():
        df_db = select_table(engine=engine, schema=schema, table_name=table_name)
        df_db['ix3'] = pd.to_datetime(df_db['ix3'])
        return df_db.set_index(index_col)

    # dtype for index for MySQL... (can't have flexible text length)
    dtype = {'ix2': VARCHAR(5)} if 'mysql' in engine.dialect.dialect_description else None

    # create
    upsert(con=engine, schema=schema, df=df_multiindex, table_name=table_name, dtype=dtype, if_row_exists='update')
    df_db = read_from_db()
    pd.testing.assert_frame_equal(df_db, df_multiindex)

    # insert
    upsert(con=engine, schema=schema, df=df_multiindex2, table_name=table_name, dtype=dtype, if_row_exists='update')
    df_db = read_from_db()
    pd.testing.assert_frame_equal(df_db, pd.concat(objs=[df_multiindex, df_multiindex2]))


@adrop_table_between_tests(table_name=TableNames.MULTIINDEX)
async def run_test_create_and_insert_table_multiindex_async(engine, schema):
    # local helpers
    table_name = TableNames.MULTIINDEX

    async def read_from_db():
        df_db = await aselect_table(engine=engine, schema=schema, table_name=table_name)
        df_db['ix3'] = pd.to_datetime(df_db['ix3'])
        return df_db.set_index(index_col)

    # dtype for index for MySQL... (can't have flexible text length)
    dtype = {'ix2': VARCHAR(5)} if 'mysql' in engine.dialect.dialect_description else None

    # create
    await aupsert(con=engine, schema=schema, df=df_multiindex, table_name=table_name,
                  dtype=dtype, if_row_exists='update')
    df_db = await read_from_db()
    pd.testing.assert_frame_equal(df_db, df_multiindex)

    # insert
    await aupsert(con=engine, schema=schema, df=df_multiindex2, table_name=table_name,
                  dtype=dtype, if_row_exists='update')
    df_db = await read_from_db()
    pd.testing.assert_frame_equal(df_db, pd.concat(objs=[df_multiindex, df_multiindex2]))


# -

# ## Test index with null value

# +
@drop_table_between_tests(table_name=TableNames.INDEX_WITH_NULL)
def run_test_index_with_null(engine, schema):
    df = pd.DataFrame({'ix': [None, 0], 'test': [0, 1]}).set_index('ix')
    with pytest.raises(IntegrityError):
        upsert(con=engine, schema=schema, df=df, table_name=TableNames.INDEX_WITH_NULL, if_row_exists='update')
        # don't test error for mysql since only a warning is raised and the line is skipped
        if 'mysql' in engine.dialect.dialect_description:  # pragma: no cover
            pytest.skip('not tested with mysql as only a warning is issued and the line is skipped. '
                        "Perhaps we could capture the warning with pytest?")


@adrop_table_between_tests(table_name=TableNames.INDEX_WITH_NULL)
async def run_test_index_with_null_async(engine, schema):
    df = pd.DataFrame({'ix': [None, 0], 'test': [0, 1]}).set_index('ix')
    with pytest.raises(IntegrityError):
        await aupsert(con=engine, schema=schema, df=df, table_name=TableNames.INDEX_WITH_NULL, if_row_exists='update')
        # don't test error for mysql since only a warning is raised and the line is skipped
        if 'mysql' in engine.dialect.dialect_description:  # pragma: no cover
            pytest.skip('not tested with mysql as only a warning is issued and the line is skipped. '
                        "Perhaps we could capture the warning with pytest?")


# -

# ## Test only index

# +
@drop_table_between_tests(table_name=TableNames.INDEX_ONLY_INSERT)
def run_test_only_index(engine, schema, if_row_exists):
    # config
    table_name = TableNames.INDEX_ONLY_INSERT

    # upsert df with only index
    df = pd.DataFrame({'ix': [1]}).set_index('ix')
    upsert(con=engine, schema=schema, df=df, table_name=table_name, if_row_exists=if_row_exists)

    # check data integrity
    df_db = select_table(engine=engine, schema=schema, table_name=table_name)
    assert 'ix' in df_db.columns
    assert len(df_db) > 0
    assert df_db['ix'].iloc[0] == 1


@adrop_table_between_tests(table_name=TableNames.INDEX_ONLY_INSERT)
async def run_test_only_index_async(engine, schema, if_row_exists):
    # config
    table_name = TableNames.INDEX_ONLY_INSERT

    # upsert df with only index
    df = pd.DataFrame({'ix': [1]}).set_index('ix')
    await aupsert(con=engine, schema=schema, df=df, table_name=table_name, if_row_exists=if_row_exists)

    # check data integrity
    df_db = await aselect_table(engine=engine, schema=schema, table_name=table_name)
    assert 'ix' in df_db.columns
    assert len(df_db) > 0
    assert df_db['ix'].iloc[0] == 1


# -

# # Actual tests

# +
def test_create_and_insert_table_multiindex(engine, schema):
    sync_or_async_test(engine=engine, schema=schema,
                       f_async=run_test_create_and_insert_table_multiindex_async,
                       f_sync=run_test_create_and_insert_table_multiindex)


def test_index_with_null(engine, schema):
    sync_or_async_test(engine=engine, schema=schema,
                       f_async=run_test_index_with_null_async,
                       f_sync=run_test_index_with_null)


@pytest.mark.parametrize("if_row_exists", ['update', 'ignore'])
def test_only_index(engine, schema, if_row_exists):
    sync_or_async_test(engine=engine, schema=schema,
                       f_async=run_test_only_index_async,
                       f_sync=run_test_only_index,
                       if_row_exists=if_row_exists)



#!/usr/bin/env python
# coding: utf-8
# +
"""
End to end test similar to the scenario proposed in the docstring
of `pangres.upsert_or_aupsert`.

We will create a table and then insert with update and then ignore for the
`ON CONFLICT` clause.
"""
import pandas as pd
import pytest
from sqlalchemy import create_engine, VARCHAR
from sqlalchemy.exc import OperationalError, ProgrammingError

# local imports
from pangres import aupsert, upsert, HasNoSchemaSystemException
from pangres.examples import _TestsExampleTable
from pangres.tests.conftest import (adrop_schema, adrop_table, adrop_table_between_tests, aselect_table,
                                    drop_schema, drop_table, drop_table_between_tests,
                                    schema_for_testing_creation, select_table,
                                    sync_async_exec_switch, sync_or_async_test, TableNames)
# -


# # Helpers

# ## Test data

# +
# df from which we will create a SQL table
df = _TestsExampleTable.create_example_df(nb_rows=6)
# test nulls won't create any problems
# avoid nulls in boolean column though as this is unpractical with pandas in older versions
df.iloc[0, [ix for ix, col in enumerate(df.columns) if col != 'likes_pizza']] = None

# df from which we will upsert_or_aupsert update or upsert_or_aupsert ignore
# remove one record from above and add one
# so we know that 1) old records are still there, 2) new ones get added
# 3) we can check whether the update/ignore of existing records worked
df2 = _TestsExampleTable.create_example_df(nb_rows=7)
df2 = df2.iloc[1:]
# -

# ## Expectations

# +
df_after_insert_update = pd.concat(objs=(df.loc[~df.index.isin(df2.index.tolist())],  # everything that is not in df2
                                         df2))

df_after_insert_ignore = pd.concat(objs=(df,
                                         df2.loc[~df2.index.isin(df.index.tolist())]))  # everything that is not in df


# -

# # Sync and async variants for tests
#
# (`run_test_foo`|`run_test_foo_async`) -> `test_foo`

# ## Upsert without DDL modifications (except `create_table`)

# +
@drop_table_between_tests(table_name=TableNames.END_TO_END)
def run_test_end_to_end(engine, schema, create_table, if_row_exists, df_expected):
    # config
    table_name = TableNames.END_TO_END
    common_kwargs_upsert = dict(if_row_exists=if_row_exists, table_name=table_name)
    read_table = lambda: _TestsExampleTable.read_from_db(engine=engine, schema=schema,
                                                         table_name=table_name).sort_index()

    # 1. create table
    upsert(con=engine, schema=schema, df=df, create_table=True, **common_kwargs_upsert)
    pd.testing.assert_frame_equal(df, read_table())

    # 2. insert update/ignore
    upsert(con=engine, schema=schema, df=df2, create_table=create_table, **common_kwargs_upsert)
    pd.testing.assert_frame_equal(df_expected, read_table())


@adrop_table_between_tests(table_name=TableNames.END_TO_END)
async def run_test_end_to_end_async(engine, schema, create_table, if_row_exists, df_expected):
    # config
    table_name = TableNames.END_TO_END
    common_kwargs_upsert = dict(if_row_exists=if_row_exists, table_name=table_name)

    async def read_table():
        temp_df = await _TestsExampleTable.aread_from_db(engine=engine, schema=schema, table_name=table_name)
        return temp_df.sort_index()

    # 1. create table
    await aupsert(con=engine, schema=schema, df=df, create_table=True, **common_kwargs_upsert)
    df_db = await read_table()
    pd.testing.assert_frame_equal(df, df_db)

    # 2. insert update/ignore
    await aupsert(con=engine, schema=schema, df=df2, create_table=create_table, **common_kwargs_upsert)
    df_db = await read_table()
    pd.testing.assert_frame_equal(df_expected, df_db)


# -

# ## Upsert with addition of new columns

# +
@drop_table_between_tests(table_name=TableNames.ADD_NEW_COLUMN)
def run_test_add_new_column(engine, schema):
    # config
    table_name = TableNames.ADD_NEW_COLUMN
    dtype = {'id': VARCHAR(5)} if 'mysql' in engine.dialect.dialect_description else None
    df = pd.DataFrame({'id': ['foo']}).set_index('id')
    # common kwargs for all the times we use upsert_or_aupsert
    common_kwargs = dict(con=engine, schema=schema, table_name=table_name,
                         if_row_exists='update', dtype=dtype)

    # 1. create table
    upsert(df=df, **common_kwargs)
    # 2. add a new column and repeat upsert_or_aupsert
    df['new_column'] = 'bar'
    upsert(df=df, add_new_columns=True, **common_kwargs)
    # verify content matches
    df_db = select_table(engine=engine, schema=schema, table_name=table_name, index_col='id').sort_index()
    pd.testing.assert_frame_equal(df, df_db)


@adrop_table_between_tests(table_name=TableNames.ADD_NEW_COLUMN)
async def run_test_add_new_column_async(engine, schema):
    # config
    table_name = TableNames.ADD_NEW_COLUMN
    dtype = {'id': VARCHAR(5)} if 'mysql' in engine.dialect.dialect_description else None
    df = pd.DataFrame({'id': ['foo']}).set_index('id')
    # common kwargs for all the times we use upsert_or_aupsert
    common_kwargs = dict(con=engine, schema=schema, table_name=table_name,
                         if_row_exists='update', dtype=dtype)

    # 1. create table
    await aupsert(df=df, **common_kwargs)
    # 2. add a new column and repeat upsert_or_aupsert
    df['new_column'] = 'bar'
    await aupsert(df=df, add_new_columns=True, **common_kwargs)
    # verify content matches
    df_db = await aselect_table(engine=engine, schema=schema, table_name=table_name, index_col='id')
    pd.testing.assert_frame_equal(df, df_db.sort_index())


# -

# ## Upsert with alteration of data type for empty columns

# +
@drop_table_between_tests(table_name=TableNames.CHANGE_EMPTY_COL_TYPE)
def run_test_adapt_column_type(engine, schema):
    # skip for sqlite as it does not support such alteration
    if 'sqlite' in engine.dialect.dialect_description:
        pytest.skip('such column alteration is not possible with SQlite')

    # config
    table_name = TableNames.CHANGE_EMPTY_COL_TYPE
    dtype = {'id': VARCHAR(5)} if 'mysql' in engine.dialect.dialect_description else None
    df = pd.DataFrame({'id': ['foo'], 'empty_column': [None]}).set_index('id')
    # common kwargs for all the times we use upsert_or_aupsert
    common_kwargs = dict(con=engine, schema=schema, df=df, table_name=table_name,
                         if_row_exists='update', dtype=dtype)

    # 1. create table
    upsert(**common_kwargs)
    # 2. add non string data in empty column and repeat upsert_or_aupsert
    df['empty_column'] = 1
    upsert(**common_kwargs, adapt_dtype_of_empty_db_columns=True)


@adrop_table_between_tests(table_name=TableNames.CHANGE_EMPTY_COL_TYPE)
async def run_test_adapt_column_type_async(engine, schema):
    # skip for sqlite as it does not support such alteration
    if 'sqlite' in engine.dialect.dialect_description:
        pytest.skip('such column alteration is not possible with SQlite')

    # config
    table_name = TableNames.CHANGE_EMPTY_COL_TYPE
    dtype = {'id': VARCHAR(5)} if 'mysql' in engine.dialect.dialect_description else None
    df = pd.DataFrame({'id': ['foo'], 'empty_column': [None]}).set_index('id')
    # common kwargs for all the times we use upsert_or_aupsert
    common_kwargs = dict(con=engine, schema=schema, df=df, table_name=table_name,
                         if_row_exists='update', dtype=dtype)

    # 1. create table
    await aupsert(**common_kwargs)
    # 2. add non string data in empty column and repeat upsert_or_aupsert
    df['empty_column'] = 1
    await aupsert(**common_kwargs, adapt_dtype_of_empty_db_columns=True)


# -

# ## Upsert with `create_schema=True` but `schema=None`

# +
@drop_table_between_tests(table_name=TableNames.CREATE_SCHEMA_NONE)
def run_test_create_schema_none(engine, schema):
    df = pd.DataFrame({'id': [0]}).set_index('id')
    upsert(con=engine, schema=None, df=df, if_row_exists='update', create_schema=True,
           table_name=TableNames.CREATE_SCHEMA_NONE, create_table=True)


@adrop_table_between_tests(table_name=TableNames.CREATE_SCHEMA_NONE)
async def run_test_create_schema_none_async(engine, schema):
    df = pd.DataFrame({'id': [0]}).set_index('id')
    await aupsert(con=engine, schema=None, df=df, if_row_exists='update', create_schema=True,
                  table_name=TableNames.CREATE_SCHEMA_NONE, create_table=True)


# -

# ## Upsert with `create_schema=True` but `schema` is not None

# +
@drop_table_between_tests(table_name=TableNames.CREATE_SCHEMA_NOT_NONE)
def run_test_create_schema_not_none(engine, schema):
    # local helpers
    is_postgres = 'postgres' in engine.dialect.dialect_description

    # overwrite schema
    schema = schema_for_testing_creation

    # config
    df = pd.DataFrame({'id': [0]}).set_index('id')
    table_name = TableNames.CREATE_SCHEMA_NOT_NONE

    # drop table before test (could not get my decorator to work with another schema
    # when having an optional arg schema=None due to variable scopes problems)
    if is_postgres:
        drop_table(engine=engine, schema=schema, table_name=table_name)
        drop_schema(engine=engine, schema=schema)

    try:
        upsert(con=engine, schema=schema, df=df, if_row_exists='update', create_schema=True,
               table_name=table_name, create_table=True)
        if not is_postgres:  # pragma: no cover
            raise AssertionError('Expected `upsert` to fail when trying to create a schema '
                                 'with another database than postgres')
    except Exception as e:
        # for postgres this should have worked
        if is_postgres:  # pragma: no cover
            raise e
        else:
            assert isinstance(e, HasNoSchemaSystemException)
    finally:
        # drop table and schema after test
        if is_postgres:
            drop_table(engine=engine, schema=schema, table_name=table_name)
            drop_schema(engine=engine, schema=schema)


@adrop_table_between_tests(table_name=TableNames.CREATE_SCHEMA_NOT_NONE)
async def run_test_create_schema_not_none_async(engine, schema):
    # local helpers
    is_postgres = 'postgres' in engine.dialect.dialect_description

    # overwrite schema
    schema = schema_for_testing_creation

    # config
    df = pd.DataFrame({'id': [0]}).set_index('id')
    table_name = TableNames.CREATE_SCHEMA_NOT_NONE

    # drop table before test (could not get my decorator to work with another schema
    # when having an optional arg schema=None due to variable scopes problems)
    if is_postgres:
        await adrop_table(engine=engine, schema=schema, table_name=table_name)
        await adrop_schema(engine=engine, schema=schema)

    try:
        await aupsert(con=engine, schema=schema, df=df, if_row_exists='update', create_schema=True,
                      table_name=table_name, create_table=True)
        if not is_postgres:  # pragma: no cover
            raise AssertionError('Expected `upsert` to fail when trying to create a schema '
                                 'with another database than postgres')
    except Exception as e:
        # for postgres this should have worked
        if is_postgres:  # pragma: no cover
            raise e
        else:
            assert isinstance(e, HasNoSchemaSystemException)
    finally:
        # drop table and schema after test
        if is_postgres:
            await adrop_table(engine=engine, schema=schema, table_name=table_name)
            await adrop_schema(engine=engine, schema=schema)


# -

# ## Test error on upsert when table does not exist (with `create_table=False`)
#
# We do not create any table here

# +
def run_test_insert_missing_table(engine, schema):
    """
    Check if an error is raised when trying to insert in a missing table
    and `create_table` is False.
    """
    df = pd.DataFrame({'id': [0]}).set_index('id')
    with pytest.raises((OperationalError, ProgrammingError)) as excinfo:
        upsert(con=engine, schema=schema, df=df, table_name=TableNames.NO_TABLE,
               if_row_exists='update', create_table=False)
    assert any(s in str(excinfo.value) for s in ('no such table', 'does not exist', "doesn't exist"))


async def run_test_insert_missing_table_async(engine, schema):
    """
    Check if an error is raised when trying to insert in a missing table
    and `create_table` is False.
    """
    df = pd.DataFrame({'id': [0]}).set_index('id')
    with pytest.raises((OperationalError, ProgrammingError)) as excinfo:
        await aupsert(con=engine, schema=schema, df=df, table_name=TableNames.NO_TABLE,
                      if_row_exists='update', create_table=False)
    assert any(s in str(excinfo.value) for s in ('no such table', 'does not exist', "doesn't exist"))


# -

# ## Test if MySQL does not automatically create an autoincremented PK when giving it integers
#
# See [issue 56](https://github.com/ThibTrip/pangres/issues/56)

# +
@drop_table_between_tests(table_name=TableNames.PK_MYSQL)
def run_test_mysql_pk_not_auto_incremented(engine, schema):
    if 'mysql' not in engine.dialect.dialect_description:
        pytest.skip('This test is only relevant for MySQL')

    table_name = TableNames.PK_MYSQL

    # upsert first df using pangres which creates the table automatically
    df1 = pd.DataFrame({'id': [0, 1], 'name': ['foo', 'bar']}).set_index('id')
    upsert(con=engine, df=df1, table_name=table_name, if_row_exists='update')

    # upsert second df
    df2 = pd.DataFrame({'id': [100, 200], 'name': ['baz', 'qux']}).set_index('id')
    upsert(con=engine, df=df2, table_name=table_name, if_row_exists='update')

    # read df back
    df_db = select_table(engine=engine, schema=schema, table_name=table_name, index_col='id')

    # check mysql got that correctly
    pd.testing.assert_frame_equal(df_db.sort_index(), pd.concat((df1, df2)).sort_index())


@adrop_table_between_tests(table_name=TableNames.PK_MYSQL)
async def run_test_mysql_pk_not_auto_incremented_async(engine, schema):
    if 'mysql' not in engine.dialect.dialect_description:
        pytest.skip('This test is only relevant for MySQL')

    table_name = TableNames.PK_MYSQL

    # upsert first df using pangres which creates the table automatically
    df1 = pd.DataFrame({'id': [0, 1], 'name': ['foo', 'bar']}).set_index('id')
    await aupsert(con=engine, df=df1, table_name=table_name, if_row_exists='update')

    # upsert second df
    df2 = pd.DataFrame({'id': [100, 200], 'name': ['baz', 'qux']}).set_index('id')
    await aupsert(con=engine, df=df2, table_name=table_name, if_row_exists='update')

    # read df back
    df_db = await aselect_table(engine=engine, schema=schema, table_name=table_name, index_col='id')

    # check mysql got that correctly
    pd.testing.assert_frame_equal(df_db.sort_index(), pd.concat((df1, df2)).sort_index())


# -

# # Actual tests

# +
# after the table is created with a first `upsert_or_aupsert`
# using `create_table=False` or `create_table=True` should both work
@pytest.mark.parametrize('create_table', [False, True], ids=['create_table_false', 'create_table_true'])
@pytest.mark.parametrize('if_row_exists, df_expected', [['update', df_after_insert_update],
                                                        ['ignore', df_after_insert_ignore]],
                         ids=['update', 'ignore'])
def test_end_to_end(engine, schema, create_table, if_row_exists, df_expected):
    sync_or_async_test(engine=engine, schema=schema,
                       f_async=run_test_end_to_end_async,
                       f_sync=run_test_end_to_end,
                       create_table=create_table,
                       if_row_exists=if_row_exists,
                       df_expected=df_expected)


@pytest.mark.parametrize('use_async', [False, True], ids=['upsert', 'aupsert'])
def test_bad_value_if_row_exists(_, use_async):
    df = pd.DataFrame({'id': [0]}).set_index('id')
    engine = create_engine('sqlite:///')
    upsert_func = upsert if use_async else aupsert
    upsert_kwargs = dict(con=engine, df=df, table_name=TableNames.NO_TABLE, if_row_exists='test')
    with pytest.raises(ValueError) as excinfo:
        sync_async_exec_switch(upsert_func, **upsert_kwargs)
    assert 'must be "ignore" or "update"' in str(excinfo.value)


def test_add_new_column(engine, schema):
    sync_or_async_test(engine=engine, schema=schema,
                       f_async=run_test_add_new_column_async,
                       f_sync=run_test_add_new_column)


def test_adapt_column_type(engine, schema):
    sync_or_async_test(engine=engine, schema=schema,
                       f_async=run_test_adapt_column_type_async,
                       f_sync=run_test_adapt_column_type)


def test_insert_missing_table(engine, schema):
    sync_or_async_test(engine=engine, schema=schema,
                       f_async=run_test_insert_missing_table_async,
                       f_sync=run_test_insert_missing_table)


def test_create_schema_none(engine, schema):
    sync_or_async_test(engine=engine, schema=schema,
                       f_async=run_test_create_schema_none_async,
                       f_sync=run_test_create_schema_none)


def test_create_schema_not_none(engine, schema):
    sync_or_async_test(engine=engine, schema=schema,
                       f_async=run_test_create_schema_not_none_async,
                       f_sync=run_test_create_schema_not_none)


def test_mysql_pk_not_auto_incremented(engine, schema):
    sync_or_async_test(engine=engine, schema=schema,
                       f_async=run_test_mysql_pk_not_auto_incremented_async,
                       f_sync=run_test_mysql_pk_not_auto_incremented)



import logging
import os
import pytest
from contextlib import contextmanager
from pangres.logger import log, loggers


# # Helpers

# +
log_level_names = ('CRITICAL', 'ERROR', 'WARNING', 'INFO', 'DEBUG')
log_level_values_to_names = {getattr(logging, k): k for k in log_level_names}


@contextmanager
def custom_logger(logger_level):
    # kick out any already initalized logger before setting the environment variable
    # not only from the `loggers` dict but also from the logging module as the handlers
    # will still be present and cause duplicates
    loggers.pop('pangres', None)
    logger = logging.getLogger('pangres')
    while logger.hasHandlers():
        try:
            logger.removeHandler(logger.handlers[0])
        except IndexError:
            break

    os.environ['PANGRES_LOG_LEVEL'] = str(logger_level)
    # make sure a logger is available (initialized if not present by the log function)
    log('DUMMY LOG', level=logging.DEBUG)
    try:
        yield
    finally:
        os.environ.pop('PANGRES_LOG_LEVEL', None)


# -

# # Tests

# +
# "_" is a dummy passed by metafunc when generating tests automatically
# for each engine and schema (see module conftests.py)
def test_bad_log_level(_):
    with pytest.raises(ValueError) as exc_info:
        log('test', level=5)
    assert 'not a valid log level' in str(exc_info.value)


# by default DEBUG is not shown
@pytest.mark.parametrize("test_level", [k for k, v in log_level_values_to_names.items() if v != 'DEBUG'],
                         scope='session')
def test_log_output(_, caplog, test_level):
    txt = 'TESTING LOG OUTPUT AND LEVEL'
    with caplog.at_level(logging.INFO):
        log(txt, level=test_level)
    assert len(caplog.records) == 1
    assert caplog.records[0].levelname == log_level_values_to_names[test_level]
    assert txt in caplog.text


def test_custom_level(_, caplog):
    base_text = 'TESTING CUSTOM LOGGER LEVEL'
    with custom_logger(logger_level=logging.WARNING):
        # test lower log level -> should be hidden
        with caplog.at_level(logging.INFO):
            log(base_text + ' | THIS LOG SHOULD BE HIDDEN', level=logging.INFO)
        assert caplog.text == ''

        # test greater than or equal level
        expected = base_text + ' | THIS LOG SHOULD BE DISPLAYED'
        with caplog.at_level(logging.INFO):
            log(expected, level=logging.WARNING)
        assert len(caplog.records) == 1
        assert expected in caplog.text



#!/usr/bin/env python
# coding: utf-8
# +
"""
This module tests we can get information back from
the upserted chunks when the parameter `yield_chunks`
is True. It also checks the integrity of the data.
"""
import math
import pandas as pd
from sqlalchemy import INT

# local imports
from pangres import aupsert, upsert
from pangres.examples import _TestsExampleTable
from pangres.tests.conftest import (adrop_table_between_tests, drop_table_between_tests,
                                    sync_or_async_test, TableNames)


# -

# # Sync and async variants for tests
#
# (`run_test_foo`|`run_test_foo_async`) -> `test_foo`

# ## Insert values one by one

# +
@drop_table_between_tests(table_name=TableNames.WITH_YIELD)
def run_test_get_nb_rows(engine, schema):
    # config
    table_name = TableNames.WITH_YIELD
    nb_rows, chunksize = 20, 3
    nb_last_chunk = nb_rows % chunksize
    nb_chunks = math.ceil(nb_rows / chunksize)
    # MySQL does not want flexible text length in indices/PK
    df = _TestsExampleTable.create_example_df(nb_rows=nb_rows)

    # iterate over upsert results
    # make sure we can extract the number of updated rows and that it is correct
    iterator = upsert(con=engine, df=df, table_name=table_name, if_row_exists='update',
                      schema=schema, chunksize=chunksize, yield_chunks=True)

    for ix, result in enumerate(iterator):
        assert result.rowcount == (chunksize if ix != (nb_chunks - 1) else nb_last_chunk)

    # verify the inserted data is as expected
    # we sort the index for MySQL
    df_db = _TestsExampleTable.read_from_db(engine=engine, schema=schema, table_name=table_name)
    pd.testing.assert_frame_equal(df.sort_index(), df_db.sort_index())


@adrop_table_between_tests(table_name=TableNames.WITH_YIELD)
async def run_test_get_nb_rows_async(engine, schema):
    # config
    table_name = TableNames.WITH_YIELD
    nb_rows, chunksize = 20, 3
    nb_last_chunk = nb_rows % chunksize
    nb_chunks = math.ceil(nb_rows / chunksize)
    # MySQL does not want flexible text length in indices/PK
    df = _TestsExampleTable.create_example_df(nb_rows=nb_rows)

    # iterate over upsert results
    # make sure we can extract the number of updated rows and that it is correct
    async_gen = await aupsert(con=engine, df=df, table_name=table_name, if_row_exists='update',
                              schema=schema, chunksize=chunksize, yield_chunks=True)

    # unlike the equivalent synchronous test, enumerate(async_generator) will not work
    ix = 0
    async for result in async_gen:
        assert result.rowcount == (chunksize if ix != (nb_chunks - 1) else nb_last_chunk)
        ix += 1

    # verify the inserted data is as expected
    # we sort the index for MySQL
    df_db = await _TestsExampleTable.aread_from_db(engine=engine, schema=schema, table_name=table_name)
    pd.testing.assert_frame_equal(df.sort_index(), df_db.sort_index())


# -

# ## Test of an empty DataFrame

# +
@drop_table_between_tests(table_name=TableNames.WITH_YIELD_EMPTY)
def run_test_yield_empty_df(engine, schema):
    df = pd.DataFrame({'id': [], 'value': []}).set_index('id')

    # we should get an empty generator back
    iterator = upsert(con=engine, df=df, table_name=TableNames.WITH_YIELD_EMPTY, if_row_exists='update',
                      schema=schema, dtype={'id': INT, 'value': INT}, yield_chunks=True)

    # the for loop should never run because the generator should be empty
    for result in iterator:
        raise AssertionError('Expected the generator returned by upsert '
                             'with an empty df to be empty')  # pragma: no cover


@adrop_table_between_tests(table_name=TableNames.WITH_YIELD_EMPTY)
async def run_test_yield_empty_df_async(engine, schema):
    df = pd.DataFrame({'id': [], 'value': []}).set_index('id')

    # we should get an empty generator back
    async_gen = await aupsert(con=engine, df=df, table_name=TableNames.WITH_YIELD_EMPTY, if_row_exists='update',
                              schema=schema, dtype={'id': INT, 'value': INT}, yield_chunks=True)

    # the for loop should never run because the generator should be empty
    async for result in async_gen:
        raise AssertionError('Expected the generator returned by aupsert '
                             'with an empty df to be empty')  # pragma: no cover


# -

# # Actual tests

# +
def test_get_nb_rows(engine, schema):
    sync_or_async_test(engine=engine, schema=schema,
                       f_async=run_test_get_nb_rows_async,
                       f_sync=run_test_get_nb_rows)


def test_yield_empty_df(engine, schema):
    sync_or_async_test(engine=engine, schema=schema,
                       f_async=run_test_yield_empty_df_async,
                       f_sync=run_test_yield_empty_df)



"""
Types and type aliases for pangres
"""
from typing import AsyncIterable, Iterable, TYPE_CHECKING, TypeVar, Union, Any
from sqlalchemy.engine import Connection, Engine
# local imports
from pangres.helpers import _sqla_gt20


if TYPE_CHECKING and _sqla_gt20():
    #  ignore F401 because some types imported here are only used in other modules
    from sqlalchemy.engine.cursor import CursorResult
    from sqlalchemy.ext.asyncio import AsyncConnection, AsyncEngine, AsyncTransaction  # noqa: F401
    from sqlalchemy.ext.asyncio.engine import AsyncConnectable  # noqa: F401
    UpsertResult = Union[None, Iterable[CursorResult]]
    AUpsertResult = Union[None, AsyncIterable[CursorResult]]
    AsyncConnectionOrAsyncEngine = Union[AsyncConnection, AsyncEngine]
else:
    # type ignore because repeated assignment is OK in this case
    AsyncConnection = TypeVar('AsyncConnection', bound=Any)  # type: ignore
    AsyncEngine = TypeVar('AsyncEngine', bound=Any)  # type: ignore
    CursorResult = TypeVar('CursorResult', bound=Any)  # type: ignore
    AsyncTransaction = TypeVar('AsyncTransaction', bound=Any)  # type: ignore
    AsyncConnectable = Union[AsyncEngine, AsyncConnection]  # type: ignore
    UpsertResult = Union[None, Iterable[CursorResult]]  # type: ignore
    AUpsertResult = Union[None, AsyncIterable[CursorResult]]  # type: ignore
    AsyncConnectionOrAsyncEngine = Union[AsyncConnection, AsyncEngine]  # type: ignore

ConnectionOrEngine = Union[Connection, Engine]
